"""Fake Preseem Model implementation."""
import time
from uuid import uuid4

import grpc
from ._apiutil import BaseModelObject
from preseem_protobuf.model.service_pb2 import SetAttachmentReq


def has_any_fields_set(message):
    for field_desc, value in message.ListFields():
        if value:
            return True
    return False


def _copy_obj(obj):
    """Copy a BaseModelObject, handling protobuf fields that can't be deepcopied."""
    new_dict = {}
    for key, value in obj.__dict__.items():
        if value is not None and hasattr(value, 'CopyFrom'):
            copy_val = type(value)()
            copy_val.CopyFrom(value)
            new_dict[key] = copy_val
        else:
            new_dict[key] = value
    return type(obj)(**new_dict)


class FakePreseemModel:
    _companies = {}  # company_uuid to Company object
    _elements = {}  # company_uuid to {element_uuid to Element object}
    _accounts = {}  # company_uuid to {account_uuid to Account object}
    _packages = {}  # company_uuid to {package_uuid to Package object}
    _services = {}  # company_uuid to {service_uuid to Service objects}
    _sites = {}  # company_uuid to {site_uuid to Site object}
    _elem_disc = {}  # company_uuid to {uuid to ElementDiscovery object}

    def _new_attrs(uuid=None):
        """Generate attributes for a new preseem-model object.
           By default this generates a random UUID, or a UUID can be passed in."""
        now = time.time()
        if not uuid:
            uuid = uuid4()
        return {
            'id': uuid.int & (1 << 64) - 1,
            'uuid': uuid.bytes,
            'created_at': now,
            'updated_at': now,
            'inactive': False,
            'inactive_time': None
        }

    def _set_inactive(obj):
        obj.inactive = True
        obj.updated_at = obj.inactive_time = time.time()

    def clear(self):
        FakePreseemModel._companies.clear()
        FakePreseemModel._elements.clear()
        FakePreseemModel._accounts.clear()
        FakePreseemModel._packages.clear()
        FakePreseemModel._services.clear()
        FakePreseemModel._sites.clear()
        FakePreseemModel._elem_disc.clear()

    async def init(self):
        pass

    async def copy_from(self, other, company_uuid):
        """Insert exact copies of entities from another model into this one."""
        try:
            for obj in await other.Company.List():
                FakePreseemModel._companies[obj.uuid] = obj
        except grpc.RpcError:  # likely permissions, fake this
            FakePreseemModel._companies[company_uuid] = other.Company(uuid=company_uuid)
        for obj in await other.Element.List(company_uuid=company_uuid):
            elems = FakePreseemModel._elements.get(company_uuid)
            if elems is None:
                elems = FakePreseemModel._elements[company_uuid] = {}
            elems[obj.uuid] = obj
        for obj in await other.Element.ListElementDiscovery(company_uuid=company_uuid):
            discs = FakePreseemModel._elem_disc.get(company_uuid)
            if discs is None:
                discs = FakePreseemModel._elem_disc[company_uuid] = {}
            discs[obj.uuid] = obj
        for obj in await other.Site.List(company_uuid=company_uuid):
            sites = FakePreseemModel._sites.get(company_uuid)
            if sites is None:
                sites = FakePreseemModel._sites[company_uuid] = {}
            sites[obj.uuid] = obj
        for obj in await other.Package.List(company_uuid=company_uuid):
            packages = FakePreseemModel._packages.get(company_uuid)
            if packages is None:
                packages = FakePreseemModel._packages[company_uuid] = {}
            packages[obj.uuid] = obj
        for obj in await other.Account.List(company_uuid=company_uuid):
            accounts = FakePreseemModel._accounts.get(company_uuid)
            if accounts is None:
                accounts = FakePreseemModel._accounts[company_uuid] = {}
            accounts[obj.uuid] = obj
        for obj in await other.Service.List(company_uuid=company_uuid,
                                            include_attachment=True):
            services = FakePreseemModel._services.get(company_uuid)
            if services is None:
                services = FakePreseemModel._services[company_uuid] = {}
            services[obj.uuid] = obj

    """Fake preseem model that implements the python API."""

    class Company(BaseModelObject):

        @classmethod
        async def Create(cls, **kwargs):
            obj = cls(**FakePreseemModel._new_attrs(kwargs.get('uuid')),
                      name=kwargs.get('name') or '',
                      parent_uuid=kwargs.get('parent_uuid'),
                      old_id=kwargs.get('old_id'))
            FakePreseemModel._companies[obj.uuid] = obj
            return _copy_obj(obj)

        @classmethod
        async def Get(cls, **kwargs):
            uuid = kwargs.get('uuid')
            assert uuid
            return _copy_obj(FakePreseemModel._companies[uuid])

        @classmethod
        async def GetByOldId(cls, **kwargs):
            old_id = kwargs.get('old_id')
            assert old_id
            companies = [
                x for x in FakePreseemModel._companies.values() if x.old_id == old_id
            ]
            assert len(companies) < 2
            if companies:
                return _copy_obj(companies[0])
            # TODO error for empty?

        @classmethod
        async def List(cls, **kwargs):
            return [_copy_obj(x) for x in FakePreseemModel._companies.values()]

    class ElementDiscovery(BaseModelObject):
        pass

    class Element(BaseModelObject):

        @classmethod
        async def Create(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            assert company_uuid
            company = FakePreseemModel._companies[company_uuid]
            obj = cls(**FakePreseemModel._new_attrs(),
                      company_uuid=company.uuid,
                      name=kwargs.get('name'),
                      serial_number=kwargs.get('serial_number'),
                      system_mac_address=kwargs.get('system_mac_address'),
                      imsi=kwargs.get('imsi'),
                      management_ip=kwargs.get('management_ip'),
                      site_uuid=kwargs.get('site_uuid'),
                      status=kwargs.get('status') or 0,
                      poller_hash=kwargs.get('poller_hash'),
                      intent_role_uuid=kwargs.get('intent_role_uuid'),
                      acting_role_uuid=kwargs.get('acting_role_uuid'),
                      source=kwargs.get('source'),
                      source_id=kwargs.get('source_id'))
            elems = FakePreseemModel._elements.get(company.uuid)
            if elems is None:
                elems = FakePreseemModel._elements[company_uuid] = {}
            elems[obj.uuid] = obj
            return _copy_obj(obj)

        @classmethod
        async def List(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            assert company_uuid
            # the real impl returns empty list if the company doesn't exist
            return [
                _copy_obj(x)
                for x in (FakePreseemModel._elements.get(company_uuid) or {}).values()
                if all([getattr(x, k) == v for k, v in kwargs.items()])
            ]

        @classmethod
        async def Delete(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            elems = FakePreseemModel._elements[company_uuid]
            if elems:
                elem = elems.get(uuid)
                if elem:
                    del elems[uuid]

        @classmethod
        async def Update(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            elems = FakePreseemModel._elements[company_uuid]
            if elems:
                elem = elems.get(uuid)
                for attr, val in kwargs.items():
                    if attr == 'company_uuid' or attr == 'uuid':
                        continue
                    elif attr == 'inactive' and val is True:
                        FakePreseemModel._set_inactive(elem)
                    else:
                        setattr(elem, attr, val)

        @classmethod
        async def CreateElementDiscovery(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            assert company_uuid
            company = FakePreseemModel._companies[company_uuid]
            obj = FakePreseemModel.ElementDiscovery(
                **FakePreseemModel._new_attrs(),
                company_uuid=company.uuid,
                name=kwargs.get('name'),
                management_ip=kwargs.get('management_ip'),
                site_uuid=kwargs.get('site_uuid'),
                source=kwargs.get('source'),
                source_id=kwargs.get('source_id'),
                intent_role_uuid=kwargs.get('intent_role_uuid'),
                element_uuid=kwargs.get('element_uuid'),
                status=kwargs.get('status'))
            discs = FakePreseemModel._elem_disc.get(company.uuid)
            if discs is None:
                discs = FakePreseemModel._elem_disc[company_uuid] = {}
            discs[obj.uuid] = obj
            return _copy_obj(obj)

        @classmethod
        async def ListElementDiscovery(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            intent_role_uuid = kwargs.get('intent_role_uuid')
            assert company_uuid
            # the real impl returns empty list if the company doesn't exist
            return [
                _copy_obj(x)
                for x in (FakePreseemModel._elem_disc.get(company_uuid) or {}).values()
                if not intent_role_uuid or intent_role_uuid == x.intent_role_uuid
            ]

        @classmethod
        async def DeleteElementDiscovery(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            discs = FakePreseemModel._elem_disc[company_uuid]
            if discs:
                disc = discs.get(uuid)
                if disc:
                    del discs[uuid]

        @classmethod
        async def UpdateElementDiscovery(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            discs = FakePreseemModel._elem_disc[company_uuid]
            if discs:
                disc = discs.get(uuid)
                for attr, val in kwargs.items():
                    if attr == 'company_uuid' or attr == 'uuid':
                        continue
                    elif attr == 'inactive' and val is True:
                        FakePreseemModel._set_inactive(disc)
                    else:
                        if getattr(disc, attr) != val:
                            disc.updated_at = time.time()
                        setattr(disc, attr, val)

    class Account(BaseModelObject):

        @classmethod
        async def Create(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            assert company_uuid
            company = FakePreseemModel._companies[company_uuid]
            obj = cls(**FakePreseemModel._new_attrs(),
                      company_uuid=company.uuid,
                      source=kwargs.get('source'),
                      source_id=kwargs.get('source_id'),
                      type=kwargs.get('type'),
                      status=kwargs.get('status'),
                      name=kwargs.get('name'),
                      first_name=kwargs.get('first_name'),
                      last_name=kwargs.get('last_name'))
            accounts = FakePreseemModel._accounts.get(company.uuid)
            if accounts is None:
                accounts = FakePreseemModel._accounts[company_uuid] = {}
            accounts[obj.uuid] = obj
            return _copy_obj(obj)

        @classmethod
        async def List(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            assert company_uuid
            # the real impl returns empty list if the company doesn't exist
            return [
                _copy_obj(x)
                for x in (FakePreseemModel._accounts.get(company_uuid) or {}).values()
            ]

        @classmethod
        async def Delete(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            accounts = FakePreseemModel._accounts[company_uuid]
            if accounts:
                account = accounts.get(uuid)
                if account:
                    del accounts[uuid]

        @classmethod
        async def Update(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            accounts = FakePreseemModel._accounts[company_uuid]
            if accounts:
                account = accounts.get(uuid)
                for attr, val in kwargs.items():
                    if attr == 'company_uuid' or attr == 'uuid':
                        continue
                    elif attr == 'inactive' and val is True:
                        FakePreseemModel._set_inactive(account)
                    else:
                        if getattr(account, attr) != val:
                            account.updated_at = time.time()
                        setattr(account, attr, val)

    class Package(BaseModelObject):

        @classmethod
        async def Create(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            assert company_uuid
            company = FakePreseemModel._companies[company_uuid]
            obj = cls(**FakePreseemModel._new_attrs(),
                      company_uuid=company.uuid,
                      name=kwargs.get('name') or '',
                      source=kwargs.get('source'),
                      source_id=kwargs.get('source_id'),
                      download_rate=kwargs.get('download_rate'),
                      upload_rate=kwargs.get('upload_rate'))
            packages = FakePreseemModel._packages.get(company.uuid)
            if packages is None:
                packages = FakePreseemModel._packages[company_uuid] = {}
            packages[obj.uuid] = obj
            return _copy_obj(obj)

        @classmethod
        async def List(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            assert company_uuid
            # the real impl returns empty list if the company doesn't exist
            return [
                _copy_obj(x)
                for x in (FakePreseemModel._packages.get(company_uuid) or {}).values()
            ]

        @classmethod
        async def Delete(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            packages = FakePreseemModel._packages[company_uuid]
            if packages:
                package = packages.get(uuid)
                if package:
                    del packages[uuid]

        @classmethod
        async def Update(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            packages = FakePreseemModel._packages[company_uuid]
            if packages:
                package = packages.get(uuid)
                for attr, val in kwargs.items():
                    if attr == 'company_uuid' or attr == 'uuid':
                        continue
                    elif attr == 'inactive' and val is True:
                        FakePreseemModel._set_inactive(package)
                    else:
                        if getattr(package, attr) != val:
                            package.updated_at = time.time()
                        setattr(package, attr, val)

    class Site(BaseModelObject):

        @classmethod
        async def Create(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            assert company_uuid
            company = FakePreseemModel._companies[company_uuid]
            obj = cls(**FakePreseemModel._new_attrs(),
                      company_uuid=company.uuid,
                      name=kwargs.get('name') or '')
            sites = FakePreseemModel._sites.get(company.uuid)
            if sites is None:
                sites = FakePreseemModel._sites[company_uuid] = {}
            sites[obj.uuid] = obj
            return _copy_obj(obj)

        @classmethod
        async def List(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            assert company_uuid
            # the real impl returns empty list if the company doesn't exist
            return [
                _copy_obj(x)
                for x in (FakePreseemModel._sites.get(company_uuid) or {}).values()
            ]

        @classmethod
        async def Delete(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            sites = FakePreseemModel._sites[company_uuid]
            if sites:
                site = sites.get(uuid)
                if site:
                    del sites[uuid]

        @classmethod
        async def Update(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            sites = FakePreseemModel._sites[company_uuid]
            if sites:
                site = sites.get(uuid)
                for attr, val in kwargs.items():
                    if attr == 'company_uuid' or attr == 'uuid':
                        continue
                    elif attr == 'inactive' and val is True:
                        FakePreseemModel._set_inactive(site)
                    else:
                        if getattr(site, attr) != val:
                            site.updated_at = time.time()
                        setattr(site, attr, val)

    class Service(BaseModelObject):

        @classmethod
        async def Create(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            assert company_uuid
            company = FakePreseemModel._companies[company_uuid]
            attachment = kwargs.get('attachment')
            obj = cls(**FakePreseemModel._new_attrs(),
                      company_uuid=company.uuid,
                      source=kwargs.get('source'),
                      source_id=kwargs.get('source_id'),
                      source_url=kwargs.get('source_url'),
                      account_uuid=kwargs.get('account_uuid'),
                      package_uuid=kwargs.get('package_uuid'),
                      download_rate=kwargs.get('download_rate'),
                      upload_rate=kwargs.get('upload_rate'),
                      attachment=attachment if has_any_fields_set(attachment) else None,
                      subscriber_identifier=kwargs.get('subscriber_identifier'))
            services = FakePreseemModel._services.get(company.uuid)
            if services is None:
                services = FakePreseemModel._services[company_uuid] = {}
            services[obj.uuid] = _copy_obj(obj)
            return _copy_obj(obj)

        @classmethod
        async def List(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            assert company_uuid
            # the real impl returns empty list if the company doesn't exist
            return [
                _copy_obj(x)
                for x in (FakePreseemModel._services.get(company_uuid) or {}).values()
            ]

        @classmethod
        async def Delete(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            services = FakePreseemModel._services[company_uuid]
            if services:
                service = services.get(uuid)
                if service:
                    del services[uuid]

        @classmethod
        async def Update(cls, **kwargs):
            company_uuid = kwargs.get('company_uuid')
            uuid = kwargs.get('uuid')
            services = FakePreseemModel._services[company_uuid]
            if services:
                service = services.get(uuid)
                for attr, val in kwargs.items():
                    if attr == 'attachment_req':
                        if val:
                            assert isinstance(val, SetAttachmentReq)
                            attr = 'attachment'
                            if has_any_fields_set(val.attachment):
                                val = val.attachment
                            else:
                                val = None
                    if attr == 'company_uuid' or attr == 'uuid':
                        continue
                    elif attr == 'inactive' and val is True:
                        FakePreseemModel._set_inactive(service)
                    else:
                        service.updated_at = time.time()
                        setattr(service, attr, val)



"""Test interval stats conversion and interpolation (STM-6045)"""
import asyncio
from hashlib import sha256
import os
import sys
import time
import unittest
from uuid import uuid4

import google.protobuf.json_format

from preseem import UpdateIntervalManager
from preseem_protobuf.network_poller.network_poller_pb2 import NetworkElementUpdate

sys.path.append(os.path.dirname(__file__))  # let code under test load stubs

def _make_poller_hash(company_uuid, mac):
    hash_id = int(mac.replace(':', ''), 16).to_bytes(6, byteorder='big')
    return sha256(company_uuid + hash_id).digest()

class TestIntervalStats(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop() # needed to initialize asyncio
        asyncio.set_event_loop(self.loop)
        self.mgr = UpdateIntervalManager()

    def tearDown(self):
        self.loop.close()

    def _await(self, co):
        return self.loop.run_until_complete(co)

    def post(self, msg, asyn=False, filt=None):
        return self._await(self._post(msg, asyn, filt))

    def test_interval_stats_01(self):
        company_uuid = uuid4().bytes
        msg = NetworkElementUpdate()
        msg.time.FromSeconds(1612927766)
        msg.company_uuid = company_uuid
        msg.instance = 'test-interval-stats'
        msg.active = msg.pingable = True

        msg.data.management_ip = '192.0.2.100'
        msg.data.name = 'Test AP'
        msg.data.system_mac_address = '0a:00:01:0b:0c:0d'
        msg.data.poller_hash = _make_poller_hash(company_uuid, msg.data.system_mac_address)
        i1 = msg.data.interfaces.add()
        i1.id = '1'
        i1.name = 'WLAN'
        i1.in_octets = 10000
        i1.in_errors = 0

        # The original message is filtered out since the interval isn't complete
        new_msg = self.mgr.handle_message(msg)
        self.assertIsNone(new_msg)

        # Send the next message to get message 1
        msg2 = NetworkElementUpdate()
        msg2.CopyFrom(msg)
        msg2.time.FromSeconds(msg.time.seconds + 60)
        msg2.data.interfaces[0].in_octets = 10060
        msg2.data.interfaces[0].in_errors = 0
        new_msg = self.mgr.handle_message(msg2)
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))
        self.assertEqual(new_msg.interval_start.seconds, 1612927740)
        self.assertEqual(new_msg.interval_end.seconds, 1612927800)

        # Send the next message to get message 2
        msg3 = NetworkElementUpdate()
        msg3.CopyFrom(msg2)
        msg3.time.FromSeconds(msg2.time.seconds + 60)
        msg3.data.interfaces[0].in_octets = 10120
        msg3.data.interfaces[0].in_errors = 0
        new_msg = self.mgr.handle_message(msg3)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 60)
        # check that counters that were 0 in both are reported as 0
        self.assertTrue(new_msg.data.interfaces[0].HasField('in_errors'))
        self.assertEqual(new_msg.data.interfaces[0].in_errors, 0)
        self.assertEqual(new_msg.interval_start.seconds, 1612927800)
        self.assertEqual(new_msg.interval_end.seconds, 1612927860)

        # The next message has an unknown/invalid value for the counter
        msg4 = NetworkElementUpdate()
        msg4.CopyFrom(msg2)
        msg4.time.FromSeconds(msg3.time.seconds + 60)
        msg4.data.interfaces[0].ClearField('in_octets')
        new_msg = self.mgr.handle_message(msg4)
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))
        self.assertEqual(new_msg.interval_start.seconds, 1612927860)
        self.assertEqual(new_msg.interval_end.seconds, 1612927920)

        # The next message results in a counter reset.
        msg5 = NetworkElementUpdate()
        msg5.CopyFrom(msg4)
        msg5.time.FromSeconds(msg4.time.seconds + 60)
        msg5.data.interfaces[0].in_octets = 10180
        new_msg = self.mgr.handle_message(msg5)
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))
        self.assertEqual(new_msg.interval_start.seconds, 1612927920)
        self.assertEqual(new_msg.interval_end.seconds, 1612927980)

        # The next message is correct and gets divided properly for a >60s delta
        msg6 = NetworkElementUpdate()
        msg6.CopyFrom(msg4)
        msg6.time.FromSeconds(msg5.time.seconds + 70)
        msg6.data.interfaces[0].in_octets = 10250
        new_msg = self.mgr.handle_message(msg6)
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))
        self.assertEqual(new_msg.interval_start.seconds, 1612927980)
        self.assertEqual(new_msg.interval_end.seconds, 1612928040)

        # The next messages is a regular increment.
        msg7 = NetworkElementUpdate()
        msg7.CopyFrom(msg6)
        msg7.time.FromSeconds(msg6.time.seconds + 60)
        msg7.data.interfaces[0].in_octets = 10320
        new_msg = self.mgr.handle_message(msg7)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 64)
        self.assertEqual(new_msg.data.interfaces[0].in_errors, 0)
        self.assertEqual(new_msg.interval_start.seconds, 1612928040)
        self.assertEqual(new_msg.interval_end.seconds, 1612928100)

        # The next message doesn't increment the counter
        msg8 = NetworkElementUpdate()
        msg8.CopyFrom(msg7)
        msg8.time.FromSeconds(msg7.time.seconds + 60)
        msg8.data.interfaces[0].in_octets = 10320
        new_msg = self.mgr.handle_message(msg8)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.interval_start.seconds, 1612928100)
        self.assertEqual(new_msg.interval_end.seconds, 1612928160)

        # The next message is a counter rollover
        msg9 = NetworkElementUpdate()
        msg9.CopyFrom(msg8)
        msg9.time.FromSeconds(msg8.time.seconds + 60)
        msg9.data.interfaces[0].in_octets = 10
        new_msg = self.mgr.handle_message(msg9)
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))
        self.assertEqual(new_msg.interval_start.seconds, 1612928160)
        self.assertEqual(new_msg.interval_end.seconds, 1612928220)

        # The next message is valid
        msg10 = NetworkElementUpdate()
        msg10.CopyFrom(msg9)
        msg10.time.FromSeconds(msg9.time.seconds + 60)
        msg10.data.interfaces[0].in_octets = 130
        new_msg = self.mgr.handle_message(msg10)
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))
        self.assertEqual(new_msg.interval_start.seconds, 1612928220)
        self.assertEqual(new_msg.interval_end.seconds, 1612928280)

        # The next message is valid
        msg11 = NetworkElementUpdate()
        msg11.CopyFrom(msg10)
        msg11.time.FromSeconds(msg10.time.seconds + 60)
        msg11.data.interfaces[0].in_octets = 250
        new_msg = self.mgr.handle_message(msg11)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 120)
        self.assertEqual(new_msg.data.interfaces[0].in_errors, 0)
        self.assertEqual(new_msg.interval_start.seconds, 1612928280)
        self.assertEqual(new_msg.interval_end.seconds, 1612928340)

        # The next message is multiple intervals
        msg12 = NetworkElementUpdate()
        msg12.CopyFrom(msg11)
        msg12.time.FromSeconds(msg11.time.seconds + 300)
        msg12.data.interfaces[0].in_octets += 600
        new_msg = self.mgr.handle_message(msg12)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 120)
        self.assertEqual(new_msg.data.interfaces[0].in_errors, 0)
        self.assertEqual(new_msg.interval_start.seconds, 1612928340)
        self.assertEqual(new_msg.interval_end.seconds, 1612928400)

        # The next message is valid
        msg13 = NetworkElementUpdate()
        msg13.CopyFrom(msg12)
        msg13.time.FromSeconds(msg12.time.seconds + 60)
        msg13.data.interfaces[0].in_octets += 120
        new_msg = self.mgr.handle_message(msg13)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 600)
        self.assertEqual(new_msg.data.interfaces[0].in_errors, 0)
        self.assertEqual(new_msg.interval_start.seconds, 1612928400)
        self.assertEqual(new_msg.interval_end.seconds, 1612928700)

        # The next message fits within the current interval
        msg14 = NetworkElementUpdate()
        msg14.CopyFrom(msg13)
        msg14.time.FromSeconds(msg13.time.seconds + 10)
        msg14.data.interfaces[0].in_octets += 20
        new_msg = self.mgr.handle_message(msg14)
        self.assertIsNone(new_msg)

        # The next message closes the interval, exactly.
        msg15 = NetworkElementUpdate()
        msg15.CopyFrom(msg14)
        msg15.time.FromSeconds(msg14.time.seconds + 14)
        msg15.data.interfaces[0].in_octets += 28
        new_msg = self.mgr.handle_message(msg15)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 120)
        self.assertEqual(new_msg.data.interfaces[0].in_errors, 0)
        self.assertEqual(new_msg.interval_start.seconds, 1612928700)
        self.assertEqual(new_msg.interval_end.seconds, 1612928760)

        # The next message is exactly one interval.
        msg16 = NetworkElementUpdate()
        msg16.CopyFrom(msg15)
        msg16.time.FromSeconds(msg15.time.seconds + 60)
        msg16.data.interfaces[0].in_octets += 83
        new_msg = self.mgr.handle_message(msg16)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 83)
        self.assertEqual(new_msg.data.interfaces[0].in_errors, 0)
        self.assertEqual(new_msg.interval_start.seconds, 1612928760)
        self.assertEqual(new_msg.interval_end.seconds, 1612928820)

        # The next message is a partial, nothing should be returned
        msg17 = NetworkElementUpdate()
        msg17.CopyFrom(msg16)
        msg17.time.FromSeconds(msg16.time.seconds + 59)
        msg17.data.interfaces[0].in_octets += 59
        new_msg = self.mgr.handle_message(msg17)
        self.assertIsNone(new_msg)

        # The next message completes that interval
        msg18 = NetworkElementUpdate()
        msg18.CopyFrom(msg17)
        msg18.time.FromSeconds(msg17.time.seconds + 60)
        msg18.data.interfaces[0].in_octets += 120
        new_msg = self.mgr.handle_message(msg18)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 61)
        self.assertEqual(new_msg.data.interfaces[0].in_errors, 0)
        self.assertEqual(new_msg.interval_start.seconds, 1612928820)
        self.assertEqual(new_msg.interval_end.seconds, 1612928880)

    def test_interval_stats_02(self):
        """Do a basic tests for all counter fields."""
        # Use two intervals of a real captured message to test this.
        with open('test/data/test_interval_stats_02_msg1.json') as f:
            json1 = f.read()
        with open('test/data/test_interval_stats_02_msg2.json') as f:
            json2 = f.read()
        with open('test/data/test_interval_stats_02_msg3.json') as f:
            json3 = f.read()
        with open('test/data/test_interval_stats_02_msg.json') as f:
            expected_result = f.read()
        msg1 = NetworkElementUpdate()
        google.protobuf.json_format.Parse(json1, msg1)
        msg2 = NetworkElementUpdate()
        google.protobuf.json_format.Parse(json2, msg2)
        msg3 = NetworkElementUpdate()
        google.protobuf.json_format.Parse(json3, msg3)
        expected_msg = NetworkElementUpdate()
        google.protobuf.json_format.Parse(expected_result, expected_msg)

        msg = self.mgr.handle_message(msg1)
        self.assertIsNone(msg)
        msg = self.mgr.handle_message(msg2)
        self.assertIsNotNone(msg)
        # The first one is all null's for the reset.
        msg = self.mgr.handle_message(msg3)
        self.assertIsNotNone(msg)
        #print(google.protobuf.json_format.MessageToJson(msg, preserving_proto_field_name=True))
        self.assertEqual(expected_msg, msg)

    def test_interval_stats_02_modules(self):
        """Do a basic tests for all counter fields for mesgs with modules."""
        # Use two intervals of a real captured message to test this.
        with open('test/data/m1_1_time20_32.json') as f:
            json1 = f.read()

        with open('test/data/m1_2_time20_36.json') as f:
            json2 = f.read()

        with open('test/data/m1_3_time20_39.json') as f:
            json3 = f.read()

        with open('test/data/m1_interval_time20_35.json') as f:
            expected_result = f.read()

        msg1 = NetworkElementUpdate()
        google.protobuf.json_format.Parse(json1, msg1)
        msg2 = NetworkElementUpdate()
        google.protobuf.json_format.Parse(json2, msg2)

        msg3 = NetworkElementUpdate()
        google.protobuf.json_format.Parse(json3, msg3)

        expected_msg = NetworkElementUpdate()
        google.protobuf.json_format.Parse(expected_result, expected_msg)

        msg = self.mgr.handle_message(msg1)
        self.assertIsNone(msg)
        msg = self.mgr.handle_message(msg2)
        self.assertIsNotNone(msg)
        # The first one is all null's for the reset.
        msg = self.mgr.handle_message(msg3)
        self.assertIsNotNone(msg)
        #print(google.protobuf.json_format.MessageToJson(msg, preserving_proto_field_name=True))
        self.assertEqual(expected_msg, msg)

    def test_interval_stats_03(self):
        """Test that an incomplete messages gets timed out.
           What do we do with a partial interval that's never completed?"""
        company_uuid = uuid4().bytes
        msg = NetworkElementUpdate()
        msg.time.FromSeconds(1612927766)
        msg.company_uuid = company_uuid
        msg.instance = 'test-interval-stats-3'
        msg.active = msg.pingable = True

        # Send three messages to get a message with a valid interval
        msg.data.management_ip = '192.0.2.100'
        msg.data.name = 'Test AP'
        msg.data.system_mac_address = '0a:00:01:0b:0c:0d'
        msg.data.poller_hash = _make_poller_hash(company_uuid, msg.data.system_mac_address)
        i1 = msg.data.interfaces.add()
        i1.id = '1'
        i1.name = 'WLAN'
        i1.out_octets = 0
        new_msg = self.mgr.handle_message(msg)
        self.assertIsNone(new_msg)

        msg2 = NetworkElementUpdate()
        msg2.CopyFrom(msg)
        msg2.time.FromSeconds(msg.time.seconds + 60)
        msg2.data.interfaces[0].out_octets += 60
        new_msg = self.mgr.handle_message(msg2)
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('out_octets'))

        msg3 = NetworkElementUpdate()
        msg3.CopyFrom(msg2)
        msg3.time.FromSeconds(msg2.time.seconds + 60)
        msg3.data.interfaces[0].out_octets += 60
        new_msg = self.mgr.handle_message(msg3)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].out_octets, 60)

        # Make sure the message doesn't expire if it's outside the expire time
        self.mgr.timeout_messages(0.1)
        msg4 = NetworkElementUpdate()
        msg4.CopyFrom(msg3)
        msg4.time.FromSeconds(msg3.time.seconds + 60)
        msg4.data.interfaces[0].out_octets += 60
        new_msg = self.mgr.handle_message(msg4)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].out_octets, 60)

        # Expire the message
        time.sleep(0.2)
        self.mgr.timeout_messages(0.1)
        msg5 = NetworkElementUpdate()
        msg5.CopyFrom(msg4)
        msg5.time.FromSeconds(msg4.time.seconds + 60)
        msg5.data.interfaces[0].out_octets += 60
        new_msg = self.mgr.handle_message(msg5)
        self.assertIsNone(new_msg)

    def test_interval_stats_04(self):
        """Test collisions between two elements, and that an id_key resolves it."""
        company_uuid = uuid4().bytes
        msg = NetworkElementUpdate()
        msg.time.FromSeconds(1612927766)
        msg.company_uuid = company_uuid
        msg.instance = 'test-interval-stats-4'
        msg.active = msg.pingable = True
        msg.data.management_ip = '192.0.2.100'
        msg.data.name = 'Test AP'
        msg.data.system_mac_address = '0a:00:01:0b:0c:0d'
        msg.data.poller_hash = _make_poller_hash(company_uuid, msg.data.system_mac_address)
        i1 = msg.data.interfaces.add()
        i1.id = '1'
        i1.name = 'WLAN'
        i1.out_octets = 0
        new_msg = self.mgr.handle_message(msg, 'idA')
        self.assertIsNone(new_msg)

        msg2 = NetworkElementUpdate()
        msg2.CopyFrom(msg)
        msg2.time.FromSeconds(msg.time.seconds + 60)
        msg2.data.interfaces[0].out_octets += 60
        new_msg = self.mgr.handle_message(msg2, 'idA')
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('out_octets'))

        # A message from another element.  Change interface name to test that
        # code path.
        msg3 = NetworkElementUpdate()
        msg3.CopyFrom(msg2)
        msg3.time.FromSeconds(msg2.time.seconds + 60)
        msg3.data.interfaces[0].out_octets = 1000
        msg3.data.interfaces[0].name = 'WLAN 1'
        new_msg = self.mgr.handle_message(msg3, 'idB')
        self.assertIsNone(new_msg)

        # Send another message for the first one to make sure its independent
        msg3 = NetworkElementUpdate()
        msg3.CopyFrom(msg2)
        msg3.time.FromSeconds(msg2.time.seconds + 60)
        msg3.data.interfaces[0].out_octets += 60
        new_msg = self.mgr.handle_message(msg3, 'idA')
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].out_octets, 60)

    def test_interval_stats_05(self):
        """A message occures within the same minute as the previous message,
           with a new station/link."""
        company_uuid = uuid4().bytes
        msg = NetworkElementUpdate()
        msg.time.FromSeconds(1612927766)
        msg.company_uuid = company_uuid
        msg.instance = 'test-interval-stats-4'
        msg.active = msg.pingable = True
        msg.data.management_ip = '192.0.2.100'
        msg.data.name = 'Test AP'
        msg.data.system_mac_address = '0a:00:01:0b:0c:0d'
        msg.data.poller_hash = _make_poller_hash(company_uuid, msg.data.system_mac_address)
        i1 = msg.data.interfaces.add()
        i1.id = '1'
        i1.name = 'WLAN'
        i1.out_octets = 0
        l1 = i1.links.add()
        l1.mac_address = '0b:00:01:00:00:01'
        l1.poller_hash = _make_poller_hash(company_uuid, l1.mac_address)
        l1.in_octets = 1000
        new_msg = self.mgr.handle_message(msg, 'idA')
        self.assertIsNone(new_msg)

        msg2 = NetworkElementUpdate()
        msg2.CopyFrom(msg)
        msg2.time.FromSeconds(msg.time.seconds + 60)
        msg2.data.interfaces[0].out_octets += 60
        msg2.data.interfaces[0].links[0].in_octets += 60
        new_msg = self.mgr.handle_message(msg2, 'idA')
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('out_octets'))
        self.assertFalse(new_msg.data.interfaces[0].links[0].HasField('out_octets'))

        msg3 = NetworkElementUpdate()
        msg3.CopyFrom(msg2)
        msg3.time.FromSeconds(msg2.time.seconds + 60)
        msg3.data.interfaces[0].out_octets += 60
        msg3.data.interfaces[0].links[0].in_octets += 60
        new_msg = self.mgr.handle_message(msg3, 'idA')
        self.assertEqual(new_msg.data.interfaces[0].out_octets, 60)
        self.assertEqual(new_msg.data.interfaces[0].links[0].in_octets, 60)

        # update within the same minute.  The way the code works now, it will
        # not add the new station until the next message.  This is an edge case.
        msg4 = NetworkElementUpdate()
        msg4.CopyFrom(msg3)
        msg4.time.FromSeconds(msg3.time.seconds + 30)
        msg4.data.interfaces[0].out_octets += 60
        msg4.data.interfaces[0].links[0].in_octets += 30
        l2 = msg4.data.interfaces[0].links.add()
        l2.mac_address = '0b:00:01:00:00:02'
        l2.poller_hash = _make_poller_hash(company_uuid, l2.mac_address)
        l2.in_octets = 1000
        new_msg = self.mgr.handle_message(msg4, 'idA')
        self.assertIsNone(new_msg)

        # back to 60s cadence
        msg5 = NetworkElementUpdate()
        msg5.CopyFrom(msg4)
        msg5.time.FromSeconds(msg4.time.seconds + 60)
        msg5.data.interfaces[0].out_octets += 120
        msg5.data.interfaces[0].links[0].in_octets += 60
        msg5.data.interfaces[0].links[1].in_octets += 60
        new_msg = self.mgr.handle_message(msg5, 'idA')
        self.assertEqual(len(new_msg.data.interfaces), 1)  # no new link yet
        self.assertEqual(new_msg.data.interfaces[0].out_octets, 94)
        self.assertEqual(new_msg.data.interfaces[0].links[0].in_octets, 60)

        msg6 = NetworkElementUpdate()
        msg6.CopyFrom(msg5)
        msg6.time.FromSeconds(msg5.time.seconds + 60)
        msg6.data.interfaces[0].out_octets += 120
        msg6.data.interfaces[0].links[0].in_octets += 60
        msg6.data.interfaces[0].links[1].in_octets += 60
        new_msg = self.mgr.handle_message(msg6, 'idA')
        self.assertEqual(new_msg.data.interfaces[0].out_octets, 120)
        self.assertEqual(new_msg.data.interfaces[0].links[0].in_octets, 60)
        self.assertEqual(new_msg.data.interfaces[0].links[1].in_octets, 60)

    def test_interval_stats_06(self):
        """A message has an interface not present in the previous message."""
        company_uuid = uuid4().bytes
        msg = NetworkElementUpdate()
        msg.time.FromSeconds(1612927766)
        msg.company_uuid = company_uuid
        msg.instance = 'test-interval-stats'
        msg.active = msg.pingable = True

        msg.data.management_ip = '192.0.2.100'
        msg.data.name = 'Test AP'
        msg.data.system_mac_address = '0a:00:01:0b:0c:0d'
        msg.data.poller_hash = _make_poller_hash(company_uuid, msg.data.system_mac_address)
        i1 = msg.data.interfaces.add()
        i1.id = '1'
        i1.name = 'WLAN'
        i1.poller_hash = _make_poller_hash(company_uuid, i1.id)
        i1.in_octets = 10000
        i1.in_errors = 0

        # The original message is filtered out since the interval isn't complete
        new_msg = self.mgr.handle_message(msg)
        self.assertIsNone(new_msg)

        # Send the next message to get message 1
        msg2 = NetworkElementUpdate()
        msg2.CopyFrom(msg)
        msg2.time.FromSeconds(msg.time.seconds + 60)
        msg2.data.interfaces[0].in_octets = 10060
        msg2.data.interfaces[0].in_errors = 0
        new_msg = self.mgr.handle_message(msg2)
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))
        self.assertEqual(new_msg.interval_start.seconds, 1612927740)
        self.assertEqual(new_msg.interval_end.seconds, 1612927800)

        # Send the next message to get message 2
        msg3 = NetworkElementUpdate()
        msg3.CopyFrom(msg2)
        msg3.time.FromSeconds(msg2.time.seconds + 60)
        msg3.data.interfaces[0].in_octets = 10120
        msg3.data.interfaces[0].in_errors = 0
        new_msg = self.mgr.handle_message(msg3)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 60)
        # check that counters that were 0 in both are reported as 0
        self.assertTrue(new_msg.data.interfaces[0].HasField('in_errors'))
        self.assertEqual(new_msg.data.interfaces[0].in_errors, 0)
        self.assertEqual(new_msg.interval_start.seconds, 1612927800)
        self.assertEqual(new_msg.interval_end.seconds, 1612927860)

        # Message 3 has a new interface
        msg4 = NetworkElementUpdate()
        msg4.CopyFrom(msg3)
        msg4.time.FromSeconds(msg3.time.seconds + 60)
        msg4.data.interfaces[0].in_octets = 10180
        msg4.data.interfaces[0].in_errors = 0
        i2 = msg4.data.interfaces.add()
        i2.id = '2'
        i2.name = 'LAN1'
        i2.poller_hash = _make_poller_hash(company_uuid, i2.id)
        i2.in_octets = 10000
        i2.in_errors = 0
        new_msg = self.mgr.handle_message(msg4)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 60)

        # Message 4 should send the new port with null counters
        msg5 = NetworkElementUpdate()
        msg5.CopyFrom(msg4)
        msg5.time.FromSeconds(msg4.time.seconds + 60)
        msg5.data.interfaces[0].in_octets = 10240
        msg5.data.interfaces[0].in_errors = 0
        msg5.data.interfaces[1].in_octets = 10120
        msg5.data.interfaces[1].in_errors = 0
        new_msg = self.mgr.handle_message(msg5)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 60)

        # Message 5 should send the new port with interval counters
        msg6 = NetworkElementUpdate()
        msg6.CopyFrom(msg5)
        msg6.time.FromSeconds(msg5.time.seconds + 60)
        msg6.data.interfaces[0].in_octets = 10300
        msg6.data.interfaces[0].in_errors = 0
        msg6.data.interfaces[1].in_octets = 10240
        msg6.data.interfaces[1].in_errors = 0
        new_msg = self.mgr.handle_message(msg6)
        self.assertIsNotNone(new_msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 60)

    def test_interval_stats_07(self):
        """STM-6177 intervals handle a one-off zero-valued counter gracefully."""
        company_uuid = uuid4().bytes
        msg = NetworkElementUpdate()
        msg.company_uuid = company_uuid
        msg.instance = 'test-interval-stats'
        msg.active = msg.pingable = True
        msg.data.management_ip = "10.16.19.2"
        msg.data.name = 'br-merom_segrain.water.gville'
        msg.data.system_mac_address = 'bc:e6:7c:20:09:0b'
        msg.data.poller_hash = _make_poller_hash(company_uuid, msg.data.system_mac_address)
        i1 = msg.data.interfaces.add()
        i1.mac_address = 'bc:e6:7c:20:09:0b'
        tmsg = msg

        msg.time.FromSeconds(1638374073) # 15:54:33
        i1.in_octets = 417705082908
        new_msg = self.mgr.handle_message(msg)
        self.assertIsNone(new_msg)

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.time.FromSeconds(1638374193) # 15:56:33
        i1 = msg.data.interfaces[0]
        i1.in_octets = 417729627827
        new_msg = self.mgr.handle_message(msg)
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.time.FromSeconds(1638374373) # 15:59:33
        i1 = msg.data.interfaces[0]
        i1.in_octets = 417744391351
        new_msg = self.mgr.handle_message(msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 21236841)

        # Send a 0.  We expect a null counter from the interval we completed,
        # because we don't know how many bytes would _really_ have been in this
        # message.
        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.time.FromSeconds(1638374493) # 16:01:33
        i1 = msg.data.interfaces[0]
        i1.in_octets = 0
        new_msg = self.mgr.handle_message(msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))

        # Continue the counter.  We do NOT want a counter reset in this case,
        # that would end up putting the entire counter value into an interval
        # bucket which is wrong in this "transient error" case where we got
        # a zero back but the counter did not, in fact, reset.
        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.time.FromSeconds(1638374613) # 16:03:33
        i1 = msg.data.interfaces[0]
        i1.in_octets = 417790449325
        new_msg = self.mgr.handle_message(msg)

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.time.FromSeconds(1638374733) # 16:05:33
        i1 = msg.data.interfaces[0]
        i1.in_octets = 417800474800
        new_msg = self.mgr.handle_message(msg)

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.time.FromSeconds(1638374853) # 16:07:33
        i1 = msg.data.interfaces[0]
        i1.in_octets = 417801806432
        new_msg = self.mgr.handle_message(msg)

    def test_interval_stats_08(self):
        """PPA-1575 changes in counter source are handled correctly."""
        company_uuid = uuid4().bytes
        msg = NetworkElementUpdate()
        msg.company_uuid = company_uuid
        msg.instance = 'test-interval-stats'
        msg.active = msg.pingable = True
        msg.data.management_ip = "10.16.19.2"
        msg.data.name = 'br-merom_segrain.water.gville'
        msg.data.system_mac_address = 'bc:e6:7c:20:09:0b'
        msg.data.poller_hash = _make_poller_hash(company_uuid, msg.data.system_mac_address)
        msg.data.counter_source = 'source1'
        i1 = msg.data.interfaces.add()
        i1.mac_address = 'bc:e6:7c:20:09:0b'
        tmsg = msg

        msg.time.FromSeconds(1638374073) # 15:54:33
        i1.in_octets = 417705082908
        new_msg = self.mgr.handle_message(msg)
        self.assertIsNone(new_msg)

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.time.FromSeconds(msg.time.seconds + 60)
        i1 = msg.data.interfaces[0]
        i1.in_octets += 60000
        new_msg = self.mgr.handle_message(msg)
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.time.FromSeconds(msg.time.seconds + 120)
        i1 = msg.data.interfaces[0]
        i1.in_octets += 180000
        new_msg = self.mgr.handle_message(msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 87000)

        # Now that we have the state going, send a message with a different counter
        # source and verify that we get no counters for the overlapping intervals.
        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.data.counter_source = 'source2'
        msg.time.FromSeconds(msg.time.seconds + 180)
        i1 = msg.data.interfaces[0]
        i1.in_octets += 240000
        new_msg = self.mgr.handle_message(msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.data.counter_source = 'source2'
        msg.time.FromSeconds(msg.time.seconds + 240)
        i1 = msg.data.interfaces[0]
        i1.in_octets += 300000
        new_msg = self.mgr.handle_message(msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.data.counter_source = 'source2'
        msg.time.FromSeconds(msg.time.seconds + 300)
        i1 = msg.data.interfaces[0]
        i1.in_octets += 360000
        new_msg = self.mgr.handle_message(msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 60000)

    def test_interval_stats_09(self):
        """PPA-1575 changes in peer counter source are handled correctly."""
        company_uuid = uuid4().bytes
        msg = NetworkElementUpdate()
        msg.company_uuid = company_uuid
        msg.instance = 'test-interval-stats'
        msg.active = msg.pingable = True
        msg.data.management_ip = "10.16.19.2"
        msg.data.name = 'br-merom_segrain.water.gville'
        msg.data.system_mac_address = 'bc:e6:7c:20:09:0b'
        msg.data.poller_hash = _make_poller_hash(company_uuid, msg.data.system_mac_address)
        i1 = msg.data.interfaces.add()
        i1.name = 'ap_wlan'
        i1.mac_address = 'bc:e6:7c:20:09:0c'
        i1.poller_hash = _make_poller_hash(company_uuid, 'bc:e6:7c:20:09:0c')
        peer = msg.peers.add()
        peer.poller_hash = _make_poller_hash(company_uuid, '00:01:02:03:04:05')
        peer.counter_source = 'source1'
        i2 = peer.interfaces.add()
        i2.name = 'sta_wlan'
        i2.poller_hash = _make_poller_hash(company_uuid, '00:01:02:03:04:06')
        l1 = i1.links.add()
        l1.poller_hash = i2.poller_hash
        l2 = i2.links.add()
        l2.poller_hash = i1.poller_hash

        tmsg = msg

        msg.time.FromSeconds(1638374073) # 15:54:33
        i1.in_octets = 100000
        i2.in_octets = 200000
        l1.in_octets = l2.out_octets = i1.in_octets
        l2.in_octets = l1.out_octets = i2.in_octets
        new_msg = self.mgr.handle_message(msg)
        self.assertIsNone(new_msg)

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.time.FromSeconds(msg.time.seconds + 60)
        i1 = msg.data.interfaces[0]
        i1.in_octets += 60000
        i2 = msg.peers[0].interfaces[0]
        i2.in_octets += 60000
        l1 = msg.data.interfaces[0].links[0]
        l2 = msg.peers[0].interfaces[0].links[0]
        l1.in_octets = l2.out_octets = i1.in_octets
        l2.in_octets = l1.out_octets = i2.in_octets
        new_msg = self.mgr.handle_message(msg)
        self.assertIsNotNone(new_msg)
        self.assertFalse(new_msg.data.interfaces[0].HasField('in_octets'))
        self.assertFalse(new_msg.data.interfaces[0].links[0].HasField('in_octets'))
        self.assertFalse(new_msg.data.interfaces[0].links[0].HasField('out_octets'))
        self.assertFalse(new_msg.peers[0].interfaces[0].HasField('in_octets'))
        self.assertFalse(new_msg.peers[0].interfaces[0].links[0].HasField('in_octets'))
        self.assertFalse(new_msg.peers[0].interfaces[0].links[0].HasField('out_octets'))

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.time.FromSeconds(msg.time.seconds + 120)
        i1 = msg.data.interfaces[0]
        i1.in_octets += 120000
        i2 = msg.peers[0].interfaces[0]
        i2.in_octets += 120000
        l1 = msg.data.interfaces[0].links[0]
        l2 = msg.peers[0].interfaces[0].links[0]
        l1.in_octets = l2.out_octets = i1.in_octets
        l2.in_octets = l1.out_octets = i2.in_octets
        new_msg = self.mgr.handle_message(msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 60000)
        self.assertEqual(new_msg.data.interfaces[0].links[0].in_octets, 60000)
        self.assertEqual(new_msg.data.interfaces[0].links[0].out_octets, 60000)
        self.assertEqual(new_msg.peers[0].interfaces[0].in_octets, 60000)
        self.assertEqual(new_msg.peers[0].interfaces[0].links[0].in_octets, 60000)
        self.assertEqual(new_msg.peers[0].interfaces[0].links[0].out_octets, 60000)

        # Now that we have the state going, send a message with a different counter
        # source and verify that we get no counters for the overlapping intervals.
        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.peers[0].counter_source = 'source2'
        msg.time.FromSeconds(msg.time.seconds + 180)
        i1 = msg.data.interfaces[0]
        i1.in_octets += 180000
        i2 = msg.peers[0].interfaces[0]
        i2.in_octets += 180000
        l1 = msg.data.interfaces[0].links[0]
        l2 = msg.peers[0].interfaces[0].links[0]
        l1.in_octets = l2.out_octets = i1.in_octets
        l2.in_octets = l1.out_octets = i2.in_octets
        new_msg = self.mgr.handle_message(msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 60000)
        self.assertFalse(new_msg.data.interfaces[0].links[0].HasField('in_octets'))
        self.assertFalse(new_msg.data.interfaces[0].links[0].HasField('out_octets'))
        self.assertFalse(new_msg.peers[0].interfaces[0].HasField('in_octets'))
        self.assertFalse(new_msg.peers[0].interfaces[0].links[0].HasField('in_octets'))
        self.assertFalse(new_msg.peers[0].interfaces[0].links[0].HasField('out_octets'))

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.peers[0].counter_source = 'source2'
        msg.time.FromSeconds(msg.time.seconds + 240)
        i1 = msg.data.interfaces[0]
        i1.in_octets += 240000
        i2 = msg.peers[0].interfaces[0]
        i2.in_octets += 240000
        l1 = msg.data.interfaces[0].links[0]
        l2 = msg.peers[0].interfaces[0].links[0]
        l1.in_octets = l2.out_octets = i1.in_octets
        l2.in_octets = l1.out_octets = i2.in_octets
        new_msg = self.mgr.handle_message(msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 60000)
        self.assertFalse(new_msg.data.interfaces[0].links[0].HasField('in_octets'))
        self.assertFalse(new_msg.data.interfaces[0].links[0].HasField('out_octets'))
        self.assertFalse(new_msg.peers[0].interfaces[0].HasField('in_octets'))
        self.assertFalse(new_msg.peers[0].interfaces[0].links[0].HasField('in_octets'))
        self.assertFalse(new_msg.peers[0].interfaces[0].links[0].HasField('out_octets'))

        msg = NetworkElementUpdate()
        msg.CopyFrom(tmsg)
        msg.peers[0].counter_source = 'source2'
        msg.time.FromSeconds(msg.time.seconds + 300)
        i1 = msg.data.interfaces[0]
        i1.in_octets += 300000
        i2 = msg.peers[0].interfaces[0]
        i2.in_octets += 300000
        l1 = msg.data.interfaces[0].links[0]
        l2 = msg.peers[0].interfaces[0].links[0]
        l1.in_octets = l2.out_octets = i1.in_octets
        l2.in_octets = l1.out_octets = i2.in_octets
        new_msg = self.mgr.handle_message(msg)
        self.assertEqual(new_msg.data.interfaces[0].in_octets, 60000)
        self.assertEqual(new_msg.data.interfaces[0].links[0].in_octets, 60000)
        self.assertEqual(new_msg.data.interfaces[0].links[0].out_octets, 60000)
        self.assertEqual(new_msg.peers[0].interfaces[0].in_octets, 60000)
        self.assertEqual(new_msg.peers[0].interfaces[0].links[0].in_octets, 60000)
        self.assertEqual(new_msg.peers[0].interfaces[0].links[0].out_octets, 60000)



import os
from typing import Dict
import yaml
import logging


def load_vendor_map() -> Dict[str, str]:
    """
    Load the YAML input file for the vendor code mapped to the vendor.
    """
    specs_filename = "vendor_map.yaml"
    dir_path = os.path.dirname(os.path.realpath(__file__))
    path = os.path.join(dir_path, specs_filename)

    # Load YAML file
    try:
        with open(path) as f:
            return yaml.safe_load(f) or {}
    except FileNotFoundError:
        logging.error(f"Vendor map file not found: {path}")
    except yaml.YAMLError as e:
        logging.error(f"Invalid YAML format in vendor map: {e}")
    except Exception as e:
        logging.error(f"Error loading vendor map: {e}")

    return {}



"""
Configuration Model.  This translates between the GRPC model and a higher-level
model for clients.
"""
import asyncio
from collections import namedtuple
from enum import Enum
import json
import logging

from google.protobuf.descriptor import FieldDescriptor
from preseem._apiutil import field_is_repeated
from google.protobuf.message_factory import GetMessageClass

from preseem.grpc_client import PreseemGrpcClient
from preseem_protobuf.config import company_configs_pb2
from preseem.pushid import generate_pushid

ConfigType = namedtuple('ConfigType', ())
ConfigOp = Enum('ConfigOp', 'ADD DEL')


def protobuf_to_dict(pb) -> dict:
    """Recursively convert a Protobuf message to a Python dict."""
    result = {}
    for field in pb.DESCRIPTOR.fields:
        value = getattr(pb, field.name)

        if field_is_repeated(field):
            # Handle repeated fields (including maps)
            if field.message_type and field.message_type.GetOptions().map_entry:
                # It's a map field
                result[field.name] = dict(value)
            else:
                # It's a repeated field (list)
                result[field.name] = [
                    protobuf_to_dict(v) if isinstance(v, Message) else v for v in value
                ]
        elif field.type == FieldDescriptor.TYPE_MESSAGE:
            # Handle nested messages
            if value is not None and value.ByteSize():
                result[field.name] = protobuf_to_dict(value)
        else:
            # Handle simple types
            result[field.name] = value

    return result


class ConfigClient(PreseemGrpcClient):

    def __init__(self, api_key, host=None, port=None, beta=None):
        super().__init__(api_key, host
                         or 'config{}.preseem.com'.format('-beta' if beta else ''),
                         port)


class ConfigModel(object):

    def __init__(self,
                 client,
                 company_id=None,
                 all_companies=None,
                 instances_only=None):
        """With a regular company-specific API key the model will act on the
           company of the API key.  With an all-companies API key, pass in
           a company_id to act as a specific company, or set all_companies=True
           to operate on all companies.  All companies operation does not
           subscribe to instances, it only processes add/delete operations
           and it calls back with (company_id, old, new) instead of the regular
           callback which just calls (old, new)."""
        self._client = client
        self.company_id = company_id  # used for multi-company keys
        self._cbk = None  # optional callback for updates
        self.all_companies = all_companies
        self.objs = {}  # public: current config objects
        self._sub_task = None  # Instance subscription
        self._sub_lock = asyncio.Lock()
        self._cfg_syn = None
        self._loaded = asyncio.Condition()
        self.instances_only = instances_only

        self.instances = {}  # current supported instances: {company -> {(type, id)}}
        self._sub_tasks = {}  # per-instance subscription state: (task, future)

        self._cfg_types = {}
        self._type_map = {}
        self._type_nt = {}
        self._type_field_is_list = {}
        self._map_types = {}
        self._nested_types = {}
        for ty, ct in self._client.config_types.items():
            for field, fd in ct.pb_type.DESCRIPTOR.fields_by_name.items():
                if fd.message_type:
                    # This is a nested message type.
                    if fd.message_type.GetOptions().map_entry:  # map type
                        value_field = fd.message_type.fields_by_name["value"]
                        self._map_types[(ty, field)] = GetMessageClass(
                            value_field.message_type
                        ) if value_field.message_type else None
                    self._nested_types[fd] = namedtuple(
                        field, [x.name for x in fd.message_type.fields])
            nt = namedtuple(ct.name, ct.pb_type.DESCRIPTOR.fields_by_name)
            self._cfg_types[ct.name.replace('_', '-')] = nt
            self._type_map[nt] = ty
            self._type_nt[ty] = nt
            self._type_field_is_list[ty] = {
                fd.name: field_is_repeated(fd)
                for fd in ct.pb_type.DESCRIPTOR.fields
            }

    def _pb(self, obj, id_prefix=None):
        """convert a config ojbect to a protobuf"""
        config_type = self._type_map.get(type(obj))
        if config_type is None:
            raise ValueError(f'Set was passed an unknown config class "{type(obj)}"')
        ct = self._client.config_types[config_type]
        d = obj._asdict()
        if not obj.id:
            d['id'] = generate_pushid()
            obj = obj._replace(id=d['id'])
        if id_prefix:
            d['id'] = f'{id_prefix}{d["id"]}'
            obj = obj._replace(id=d['id'])
        if obj.extra:
            d['extra'] = json.dumps(obj.extra)
        map_vars = {}
        for map_var in [v for v in obj._fields if (config_type, v) in self._map_types]:
            map_vars[map_var] = d.pop(map_var)
        pb = ct.pb_type(**d)
        for k, v in map_vars.items():
            if v:
                valtype = self._map_types[(config_type, k)]
                if valtype:
                    for kk, vv in v.items():
                        getattr(pb, k)[kk].CopyFrom(valtype(**vv))
                else:
                    setattr(pb, k, v)
        return pb

    def _pb_to_nt(self, config_type, pb):
        """Convert a config protobuf to a namedtuple object"""

        def pythonize_field(pb, f):
            if f == 'extra':
                return extra
            val = getattr(pb, f, None)
            if field_is_list.get(f) and (config_type, f) not in self._map_types:
                val = list(val)
            elif f in nested_fields:
                if (config_type, f) in self._map_types:
                    mapval = self._map_types.get((config_type, f))
                    if mapval:  # value is a protobuf type
                        val = {x: protobuf_to_dict(y) for x, y in val.items()}
                    else:  # value is a simple type
                        val = {x: y for x, y, in val.items()}
                else:
                    val = {
                        x.name: getattr(val, x.name)
                        for x in nested_fields[f].message_type.fields
                    }

            return val

        nt = self._type_nt.get(config_type)
        if nt is None:
            raise ValueError(f'Unsupported config type {config_type}')
        extra = None
        if pb.extra:
            try:
                extra = json.loads(pb.extra)
            except AttributeError:
                extra = {}
            except json.JSONDecodeError:
                logging.warning('Bad JSON in extra field for %s: "%s"', config_type,
                                pb.extra)
            if not isinstance(extra, dict):
                logging.warning('extra field is not a dict for %s: "%s"', config_type,
                                pb.extra)
        nested_fields = {
            x.name: x
            for x in pb.DESCRIPTOR.fields if x in self._nested_types
        }
        field_is_list = self._type_field_is_list[config_type]
        return nt(*[pythonize_field(pb, f) for f in nt._fields])

    def close(self):
        self._client.close_subscriptions()
        if self._sub_task:
            self._sub_task.cancel()
            self._sub_task = None
        for task, fut in self._sub_tasks.values():
            if fut:
                fut.cancel()
            task.cancel()
        self._sub_tasks.clear()

    def dict2cfg(self, ty, d):
        """Convert a dict to a config object."""
        nt = self._cfg_types.get(ty)
        if nt is None:
            raise RuntimeError('Unknown config type "%s"', ty)
        spec = {x: None for x in nt._fields}
        spec.update(d)
        return nt(**spec)

    def get_type(self, name):
        """Return a generated python type for the named config type."""
        return self._cfg_types.get(name)

    def get_default(self, name):
        """Return an "empty" config object for the named config type."""
        nt = self.get_type(name)
        return self._pb_to_nt(self._type_map[nt],
                              self._pb(nt(*[None for _ in nt._fields])))

    async def delete(self, obj, company_id=None):
        """Delete a configuration object."""
        config_type = self._type_map.get(type(obj))
        if config_type is None:
            raise ValueError(f'Set was passed an unknown config class "{type(obj)}"')
        ct = self._client.config_types[config_type]
        r = await self._client.del_config(config_type, obj.id, company_id
                                          or self.company_id)
        if not r:
            raise RuntimeError('Failed to delete config instance "{}"'.format(obj.id))

    async def list(self):
        """Return a list of config objects."""
        loop = asyncio.get_event_loop()
        fut = loop.create_future()
        result = []

        async def cbk(msg, syn=False):
            if msg.event_type == company_configs_pb2.CompanyConfigEvent.LOADED:
                fut.set_result(True)
            else:
                try:
                    r = await self._client.get_config(msg.config_type, msg.config_id,
                                                      msg.company_id)
                    result.append(self._pb_to_nt(msg.config_type, r))
                except ValueError:
                    pass  # we don't know about this type, skip it

        cor = self._client.sub_company_configs(cbk,
                                               company_id=self.company_id,
                                               subscribe_all=self.all_companies)
        task = loop.create_task(cor)
        try:
            await asyncio.wait_for(fut, timeout=10)
        finally:
            task.cancel()
        return result

    async def set(self, obj, company_id=None, id_prefix=None):
        """Set a configuration object."""
        config_type = self._type_map.get(type(obj))
        pb = self._pb(obj, id_prefix=id_prefix)
        r = await self._client.set_config(config_type, pb, company_id
                                          or self.company_id)
        if not r:
            raise RuntimeException('Failed to set config instance "{}"'.format(obj))
        return obj._replace(id=pb.id)

    async def handle_event(self, msg, syn=False):
        """
        Handle a config instance event from the event stream.
        The end result of this is an up to date set of config objects.
        The first time through, call the update() method to wait for everything
        to be fully loaded.  After that, "objs" always has a consistent state.
        This can also call a callback whenever something changes.
        """
        if syn:
            self._cfg_syn = {}  # Keep track of the initial stream of instances

        if msg.event_type == company_configs_pb2.CompanyConfigEvent.LOADED:
            self.instances = self._cfg_syn or {}
            self._cfg_syn = None
            # wait for any instance subscriptions to be initialized
            pend = [fut for _, fut in self._sub_tasks.values() if fut]
            if pend:
                done, pend = await asyncio.wait(pend)
                [x.result() for x in done]
            if self.instances_only and self._cbk:
                for company_id, instances in self.instances.items():
                    for config_type, config_id in instances:
                        try:
                            await self._cbk(ConfigOp.ADD, company_id, config_type,
                                            config_id)
                        except asyncio.CancelledError:
                            raise
                        except Exception as err:
                            logging.warning('Error with initial callback: %s', err)
            async with self._loaded:
                self._loaded.notify()
            return
        elif msg.event_type == company_configs_pb2.CompanyConfigEvent.UPDATE:
            return  # we don't care about modify events at the company level
        elif msg.config_type not in self._client.config_types:
            return  # we don't know/care about this config type

        loop = asyncio.get_event_loop()
        uid = (msg.company_id, msg.config_type, msg.config_id)
        if msg.event_type == company_configs_pb2.CompanyConfigEvent.SET:
            op = ConfigOp.ADD
            if self._cfg_syn is None:
                instances = self.instances.get(msg.company_id)
                if instances is None:
                    instances = self.instances[msg.company_id] = set()
                instances.add((msg.config_type, msg.config_id))
            else:
                instances = self._cfg_syn.get(msg.company_id)
                if instances is None:
                    instances = self._cfg_syn[msg.company_id] = set()
                instances.add((msg.config_type, msg.config_id))
            if not self.instances_only:
                sub_state = self._sub_tasks.get(uid)
                if sub_state:  # unexpected, race condition?
                    pass
                else:
                    # subscribe to the instance
                    task = loop.create_task(
                        self._client.sub_config(msg.config_type, msg.config_id,
                                                self.handle_instance_update,
                                                msg.company_id))
                    sub_state = self._sub_tasks[uid] = (task, loop.create_future())

        elif msg.event_type == company_configs_pb2.CompanyConfigEvent.DELETE:
            op = ConfigOp.DEL
            instances = self.instances.get(msg.company_id)
            if instances:
                instances.discard((msg.config_type, msg.config_id))
                if not instances:
                    del self.instances[msg.company_id]
            if not self.instances_only:
                # stop subscription and callback for instance removal
                sub_state = self._sub_tasks.get(uid)
                if sub_state:
                    task, fut = sub_state
                    if fut:
                        fut.cancel()
                    task.cancel()
                    del self._sub_tasks[uid]

                cur = self.objs.get(msg.config_id)
                if cur:
                    del self.objs[msg.config_id]
                    if self._cbk:  # send a deletion callback
                        try:
                            await self._cbk(cur, None)
                        except asyncio.CancelledError:
                            raise
                        except Exception as err:
                            logging.warning('Error deleting config "%s": %s',
                                            msg.config_id, err)

        if self.instances_only and self._cbk and self._cfg_syn is None:
            # Call a special callback for instance operations only
            try:
                await self._cbk(op, msg.company_id, msg.config_type, msg.config_id)
            except asyncio.CancelledError:
                raise
            except Exception as err:
                logging.warning('Error calling instances callback: %s', err)

    async def handle_instance_update(self, company_id, config_type, msg):
        """Handle a change to a config model object."""
        cur = self.objs.get(msg.id)
        obj = self._pb_to_nt(config_type, msg)
        if cur == obj:
            logging.info('Ignoring config update of "%s" with no change', msg.id)
            return
        self.objs[msg.id] = obj
        if self._cbk:
            try:
                await self._cbk(cur, obj)
            except Exception as err:
                logging.warning('Error updating "%s": %s', msg.id, err)
        # notify the initial company subscription if it is waiting
        sub_state = self._sub_tasks.get((company_id, config_type, msg.id))
        if sub_state and sub_state[1] and not sub_state[1].done():
            sub_state[1].set_result(True)

    async def listen(self, cbk):
        """Listen for events on config objects (in the background)."""
        if self._sub_task is None:
            async with self._sub_lock:
                if self._sub_task is None:
                    self._cbk = cbk
                    self._sub_task = asyncio.get_event_loop().create_task(
                        self._client.sub_company_configs(
                            self.handle_event,
                            company_id=self.company_id,
                            subscribe_all=self.all_companies))
                    async with self._loaded:
                        await self._loaded.wait()



import unittest
import os
import random
import string

from preseem.ringbuffer import RingBuffer, SHARED_MEMORY_AVAILABLE

@unittest.skipIf(not SHARED_MEMORY_AVAILABLE, "RingBuffer requires Python 3.8+ (shared_memory not available)")
class TestRingBuffer(unittest.TestCase):

    def setUp(self):
        """
        Create a temporary ring buffer for each test,
        so tests are isolated and do not interfere.
        """
        # Generate a random shared memory name to avoid collisions
        self.buffer_name = "test_ring_" + "".join(random.choices(string.ascii_lowercase, k=8))
        self.capacity = 256
        self.ring = RingBuffer(name=self.buffer_name, capacity=self.capacity, create=True)

    def tearDown(self):
        """
        Clean up shared memory at the end of each test.
        """
        self.ring.close()
        self.ring.unlink()

    def test_empty_read_returns_none(self):
        """Reading from an empty buffer should return None."""
        self.assertIsNone(self.ring._read(), "Expected None when reading from empty ring buffer")

    def test_single_message(self):
        """Write a single message, then read it back."""
        msg = b"Hello, World!"
        self.ring._write(msg)
        read_back = self.ring._read()
        self.assertEqual(read_back, msg, "Message read should match the message written")

    def test_multiple_small_messages(self):
        """Write multiple small messages and verify read order."""
        messages = [b"msg1", b"msg2", b"msg3"]
        for m in messages:
            self.ring._write(m)

        for expected in messages:
            actual = self.ring._read()
            self.assertEqual(actual, expected)

        # Ring should be empty now
        self.assertIsNone(self.ring._read())

    def test_wraparound_write(self):
        """
        Force the buffer to wrap around by writing enough data so
        the tail crosses the physical end, then verify messages.
        """
        # We'll use a small capacity (256) to force wrap
        msg1 = b"A" * 100   # 100 bytes
        msg2 = b"B" * 100   # Another 100 bytes
        msg3 = b"C" * 40    # 40 bytes => total ~240 bytes used

        self.ring._write(msg1)
        self.ring._write(msg2)
        self.ring._write(msg3)

        # Now reading them should still give us the correct content:
        self.assertEqual(self.ring._read(), msg1)
        self.assertEqual(self.ring._read(), msg2)
        self.assertEqual(self.ring._read(), msg3)
        self.assertIsNone(self.ring._read(), "Buffer should be empty after reading all messages")

    def test_discarding_old_messages(self):
        """
        Ensure that old messages are discarded when there's not enough room.
        We'll create a scenario where multiple old messages must be discarded.
        """
        # Fill buffer nearly to capacity with multiple small messages
        msg_size = 30
        count = (self.capacity // (msg_size + 4))  # each message has 4-byte prefix
        for i in range(count):
            data = (f"msg{i}").encode().ljust(msg_size, b"x")
            self.ring._write(data)

        # Now attempt to write a larger message that won't fit by discarding just one old message
        large_msg = b"L" * 100
        self.ring._write(large_msg)  # Should discard enough old messages

        # Read until we find our large message
        found_large_msg = False
        while True:
            result = self.ring._read()
            if result is None:
                break
            if result == large_msg:
                found_large_msg = True

        self.assertTrue(found_large_msg, "Failed to find the large message; discard logic might be incorrect")

    def test_message_too_large(self):
        """Try writing a message larger than the entire capacity."""
        too_big = b"A" * (self.capacity + 1)
        with self.assertRaises(ValueError):
            self.ring._write(too_big)

    def test_corrupted_message_size_detection(self):
        """Test that reading corrupted message size is detected."""
        import struct
        # Write a valid message first
        msg = b"valid message"
        self.ring._write(msg)

        # Manually corrupt the message size prefix to an invalid huge value
        # The prefix is at META_SIZE + head (which is 0)
        self.ring._attach()
        corrupt_size = self.capacity + 1000  # Larger than capacity
        struct.pack_into('>I', self.ring.shm.buf, self.ring.META_SIZE, corrupt_size)

        # Reading should now raise ValueError
        with self.assertRaises(ValueError, msg="Should detect corrupted message size"):
            self.ring._read()

    def test_zero_message_size_detection(self):
        """Test that zero message size is detected as corruption."""
        import struct
        # Write a valid message first
        msg = b"valid message"
        self.ring._write(msg)

        # Manually corrupt the message size prefix to zero
        self.ring._attach()
        struct.pack_into('>I', self.ring.shm.buf, self.ring.META_SIZE, 0)

        # Reading should now raise ValueError
        with self.assertRaises(ValueError, msg="Should detect zero message size"):
            self.ring._read()



"""
Fake Mikrotik API Client for testing and for reading recorded commands.
"""
from .mikrotik import MikrotikApiClient
import yaml


class FakeMikrotikApiClient(MikrotikApiClient):

    def __init__(self,
                 host,
                 username,
                 password,
                 enable_ssl=False,
                 port=None,
                 timeout=30,
                 ignore_permissions=False,
                 ignore_internal_errors=False,
                 record=False):
        self.commands = []  # set to [[words], [results]]
        super().__init__('TEST',
                         username,
                         password,
                         ignore_permissions=ignore_permissions,
                         ignore_internal_errors=ignore_internal_errors)
        infile = host
        if infile:
            with open(infile) as f:
                self.commands = yaml.safe_load(f.read())['commands']
        self._cmd = None
        self._buf = None

    async def connect(self):
        pass

    async def read(self):
        if not self._buf:
            try:
                self._buf = self.commands.pop([x[0] for x in self.commands
                                               ].index(self._cmd))[1]
            except ValueError:
                # if we didn't capture this command, just return empty result
                return ['!done']
        return self._buf.pop(0)

    async def write(self, words):
        self._cmd = words



def normalize_manufacturer(manufacturer):
    """ Normalize a manufacturer name to a specific format """
    if manufacturer and len(manufacturer) > 0:
        manufacturer = manufacturer.strip()  # Remove leading and trailing whitespace
        # Verify and reassign to correct manufacturer names
        if "fs" == manufacturer.lower():
            manufacturer = "FS"
        elif "fiberstore" == manufacturer.lower():
            manufacturer = "FiberStore"
        elif "brocade" == manufacturer.lower():
            manufacturer = "BROCADE"
        # TODO: To be gradually uncommented when required in the future
        # elif "waverider" == manufacturer.lower():
        #     manufacturer = "WaveRider"
        # elif "bdcom" == manufacturer.lower():
        #     manufacturer = "BDCOM"
        # elif "cmts" == manufacturer.lower():
        #     manufacturer = "CMTS"
        # elif "ignitenet" == manufacturer.lower():
        #     manufacturer = "IgniteNet"
        # elif "staros" == manufacturer.lower():
        #     manufacturer = "StarOS"
        # elif "vsol" == manufacturer.lower():
        #     manufacturer = "VSOL"
        # elif "alfocom" == manufacturer.lower():
        #     manufacturer = "ALFOCOM"
        # elif "ligowave" == manufacturer.lower():
        #     manufacturer = "LigoWave"
        # elif "oem" == manufacturer.lower():
        #     manufacturer = "OEM"
        # elif "wtd" == manufacturer.lower():
        #     manufacturer = "WTD"
        # elif "linktel" == manufacturer.lower():
        #     manufacturer = "LINKTEL"
        # elif "us" == manufacturer.lower():
        #     manufacturer = "US"
        # elif "flexoptix" == manufacturer.lower():
        #     manufacturer = "FLEXOPTIX"
        else:
            # Capitalize only first letter
            manufacturer = manufacturer.lower()
            manufacturer = manufacturer[0].upper() + manufacturer[1:]
    return manufacturer



"""
Tests for get_brmacs functionality in http_client.
"""

import asyncio
import unittest
from preseem.ssh_client import SshClient, NoBridgeError


class FakeSshClient(SshClient):
    """Stub out SSH Client."""

    async def scmd(self, cmd):
        """Fake scmd method to stub the ssh querying."""
        showmacs_output =  \
           """port no mac addr                is local?       ageing timer
                2     60:22:32:39:4d:32       no                 0.51
                1     60:22:32:6f:ab:a0       yes                0.00
                1     60:22:32:6f:ab:a0       yes                0.00
                2     62:22:32:39:4d:32       yes                0.00
                2     62:22:32:39:4d:32       yes                0.00
                1     c4:ad:34:2c:e3:6e       no                 0.05
                2     dc:2c:6e:c3:6d:88       no                 0.00
                2     dc:2c:6e:d1:76:cd       no                 1.43"""
        showmacs_output_br2 =  \
           """port no mac addr                is local?       ageing timer
                1     60:22:32:6f:ab:a0       yes                0.00
                1     60:22:32:6f:ab:a0       yes                0.00
                2     62:22:32:39:4d:32       yes                0.00
                2     62:22:32:39:4d:32       yes                0.00"""
        cmd = cmd.split()
        if cmd[0] == 'brctl' and len(cmd) > 2:
            if cmd[1] == 'showstp':
                if cmd[2] in ('br0', 'br2'):
                    return 'eth0 (1)\nprs0 (2)'
                else:
                    return ''
            elif cmd[1] == 'showmacs':
                if cmd[2] == 'br0':
                    return showmacs_output
                if cmd[2] == 'br2':
                    return showmacs_output_br2


class TestSshClient(unittest.TestCase):
    """Tests for SSH client."""

    def setUp(self):
        self.loop = asyncio.new_event_loop()  # needed to initialize asyncio
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.loop.close()

    def _await(self, co):
        return self.loop.run_until_complete(co)

    def test_get_brmacs(self):
        """Test functionality of get_brmacs for ssh_client."""
        # Test valid brname
        client = FakeSshClient('127.0.0.1')
        brmacs_results = self._await(client.get_brmacs('br0'))
        assert (brmacs_results)
        eth_macs = brmacs_results.get('eth0')
        self.assertEqual(eth_macs[0], 'c4:ad:34:2c:e3:6e')

        # Test invalid brname
        with self.assertRaises(NoBridgeError):
            brmacs_results = self._await(client.get_brmacs('br1'))

        # Test case with all local mac_address in the showmacs output
        brmacs_results = self._await(client.get_brmacs('br2'))
        assert (brmacs_results)
        eth_macs = brmacs_results.get('eth0')
        self.assertEqual(len(eth_macs), 0)



import asyncio
import concurrent.futures
import queue
import threading
import time
import unittest

import preseem.grpc_nm_model
from preseem.grpc_nm_model import Op
from preseem import PreseemNetworkMetadataModel, NetworkMetadataEntity, NetworkMetadataIndex, NetworkMetadataLowercaseIndex, NetworkMetadataReference, Reference
from preseem_grpc_model import network_metadata_pb2

class _Disconnected(Exception):
    """Simulates a gRPC disconnect to stop the iterator."""
    pass

_SENTINEL = object()

class GrpcIterator:
    """Simulates a gRPC subscription iterator with a .next() method."""
    def __init__(self, messages=None):
        self._q = queue.Queue()
        for msg in (messages or []):
            self._q.put(msg)
        self._q.put(_SENTINEL)

    def put(self, item):
        self._q.put(item)

    def stop(self):
        self._q.put(_SENTINEL)

    def next(self):
        item = self._q.get()
        if item is _SENTINEL:
            raise _Disconnected
        return item


def _consume_handler(model, it):
    """Run ref_handler, collecting ops until the iterator is exhausted."""
    ops = []
    try:
        for op in model.ref_handler(it):
            ops.append(op)
    except _Disconnected:
        pass
    return ops


def _make_entity(net):
    """Helper to make a NetworkMetadataEntity in a standard way."""
    return NetworkMetadataEntity(net, time.time(), {
        'session': Reference(net),
        'snmp_session': Reference(net),
    })

class FakeMetadataClient(object):
    """Fake nm client"""
    def __init__(self):
        self.nms = []  # entities to set in the fake server
        self.refs = []  # references to set in the fake server

    def close_subscriptions(self):
        pass

    async def populate_ents(self):
        """Fake sending entities via a subscription."""
        syn = True
        async def send_msg(ent):
            """Send a message.  If ent is "None", send the LOADED action."""
            nonlocal syn
            await self.send_ent_entity_action(ent, syn=syn)
            syn = False

        for ent in self.nms:
            await send_msg(ent)
        await send_msg(None)
        self.done_fut.set_result(None)

    async def send_ent_entity_action(self, ent, delete=False, syn=False, fut=None):
        """Send a message to the subscription.  If ent is None, send the LOADED action."""
        ts = network_metadata_pb2.google_dot_protobuf_dot_timestamp__pb2.Timestamp()             
        ts.FromNanoseconds(int(time.time() * 1000000000))
        action = network_metadata_pb2.NetworkMetadataEntityAction.SET
        set_ent = None
        if ent:
            set_ent = network_metadata_pb2.NetworkMetadataEntity(address_space='default', network_prefix=ent.network_prefix, start=ts)
            preseem.grpc_nm_model._nm_set_attrs(set_ent.attributes, ent.attributes)
            if delete:
                action = network_metadata_pb2.NetworkMetadataEntityAction.DELETE
        else:
            action = network_metadata_pb2.NetworkMetadataEntityAction.LOADED
        msg = network_metadata_pb2.NetworkMetadataEntityAction(action=action, entity=set_ent, timestamp=ts)
        await self.ent_cbk(msg, syn)
        if fut:
            fut.set_result(None)

    async def send_ref_entity_action(self, ref, delete=False, syn=False, fut=None):
        """Send a ref update using the new Op-based callback signature."""
        if ref is None:
            # LOADED — not used for individual sends, only in populate_refs
            pass
        elif delete:
            await self.ref_cbk(Op.DEL, ref)
        else:
            await self.ref_cbk(Op.SET, ref)
        if fut:
            fut.set_result(None)

    def get_ref_entity_action(self, ref, delete=False):
        """Get a subscription message.  If ref is None, send the LOADED action."""
        ts = network_metadata_pb2.google_dot_protobuf_dot_timestamp__pb2.Timestamp()             
        ts.FromNanoseconds(int(time.time() * 1000000000))
        action = network_metadata_pb2.NetworkMetadataReferenceEntityAction.SET
        set_ref = None
        if ref:
            set_ref = network_metadata_pb2.NetworkMetadataReferenceEntity(type=ref.type, value=ref.value)
            preseem.grpc_nm_model._nm_set_attrs(set_ref.attributes, ref.attributes)
            if delete:
                action = network_metadata_pb2.NetworkMetadataReferenceEntityAction.DELETE
        else:
            action = network_metadata_pb2.NetworkMetadataReferenceEntityAction.LOADED
        msg = network_metadata_pb2.NetworkMetadataReferenceEntityAction(action=action, entity=set_ref, timestamp=ts)
        return msg

    async def del_nm_async(self, ent):
        """HAve an entity asynchronously deleted from the subscription."""
        loop = asyncio.get_running_loop()
        fut = loop.create_future()
        loop.create_task(self.send_ent_entity_action(ent, delete=True, fut=fut))
        await fut

    async def set_nm_async(self, ent):
        """HAve an entity asynchronously deleted from the subscription."""
        loop = asyncio.get_running_loop()
        fut = loop.create_future()
        loop.create_task(self.send_ent_entity_action(ent, fut=fut))
        await fut

    async def del_ref_async(self, ref):
        """Have a reference asynchronously deleted from the subscription."""
        loop = asyncio.get_running_loop()
        fut = loop.create_future()
        loop.create_task(self.send_ref_entity_action(ref, delete=True, fut=fut))
        await fut

    async def set_ref_async(self, ref):
        """Have a reference asynchronously set into the subscription."""
        loop = asyncio.get_running_loop()
        fut = loop.create_future()
        loop.create_task(self.send_ref_entity_action(ref, fut=fut))
        await fut

    async def populate_refs(self, handler):
        """Fake sending refs via a subscription, using the new handler/thread pattern."""
        messages = [self.get_ref_entity_action(ref) for ref in self.refs] + \
                   [self.get_ref_entity_action(None)]
        it = GrpcIterator(messages)
        loop = asyncio.get_running_loop()

        def worker():
            try:
                for op in handler(it):
                    try:
                        fut = asyncio.run_coroutine_threadsafe(self.ref_cbk(*op), loop)
                        fut.result()
                    except concurrent.futures.CancelledError:
                        return
            except _Disconnected:
                pass  # iterator exhausted

        await loop.run_in_executor(None, worker)
        if not self.done_fut.done():
            self.done_fut.set_result(None)

    async def nm_ent_subscribe(self, cbk, company_id=None):
        self.ent_cbk = cbk
        loop = asyncio.get_running_loop()
        self.done_fut = loop.create_future()
        loop.create_task(self.populate_ents())
        return self.done_fut

    async def nm_ref_subscribe(self, cbk, company_id=None, handler=None):
        self.ref_cbk = cbk
        loop = asyncio.get_running_loop()
        self.done_fut = loop.create_future()
        loop.create_task(self.populate_refs(handler))
        return self.done_fut

class TestNetworkMetadataModel(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop() # needed to initialize asyncio
        asyncio.set_event_loop(self.loop)
        self.client = FakeMetadataClient()
        self.model = PreseemNetworkMetadataModel(self.client, refs_by_type=True)

    def tearDown(self):
        self.model.close()
        self.loop.run_until_complete(self.loop.shutdown_asyncgens())
        self.loop.close()

    def update(self):
        return self.loop.run_until_complete(self.model.update())

    def _wait_for_queue(self):
        """Wait for the flex model queue to be emptied."""
        self.loop.run_until_complete(self.model._flex_q.join())

    def _wait(self, sec):
        self.loop.run_until_complete(asyncio.sleep(sec))

    def check_req(self, pb, flexes):
        """Verify that a list of flexes is the same as a flex request."""
        req = metrics_pb2.FlexPushRequest()
        req.flexes.extend([self.model._flex_pb(x) for x in flexes])
        self.assertEqual(pb, req)

    def del_nm(self, ent):
        self.loop.run_until_complete(self.client.del_nm_async(ent))

    def set_nm(self, ent):
        self.loop.run_until_complete(self.client.set_nm_async(ent))

    def del_ref(self, ref):
        self.loop.run_until_complete(self.client.del_ref_async(ref))

    def set_ref(self, ref):
        self.loop.run_until_complete(self.client.set_ref_async(ref))

    def test_lowercase_index(self):
        """Test lowercase reference index"""
        # Load fake refs
        self.model.close()
        indexes = [NetworkMetadataLowercaseIndex('service', 'username'),]
        self.model = PreseemNetworkMetadataModel(self.client, refs_by_type=True, indexes=indexes)
        self.client.refs = [
            NetworkMetadataReference('service', 'service1', {}),
            NetworkMetadataReference('service', 'service2', {'username': 'User1'})
        ]
        self.update()

        # Lookup from initial load
        refs = self.model.get_ref_index('service', 'username', 'user1') or {}
        self.assertEqual(len(refs), 1)
        self.assertEqual(refs.get('service2', None), self.client.refs[1])

        # Load in a new ref to update the index
        new_ref = NetworkMetadataReference('service', 'service8', {'username': 'User8'})
        self.set_ref(new_ref)
        refs = self.model.get_ref_index('service', 'username', 'user8') or {}
        self.assertEqual(len(refs), 1)
        self.assertEqual(refs.get('service8', None), new_ref)

        # Delete a ref and make sure it's removed from the index
        self.del_ref(new_ref)
        refs = self.model.get_ref_index('service', 'username', 'user8') or {}
        self.assertEqual(len(refs), 0)

    def test_nm_trie(self):
        # Load fake entities
        self.model.close()
        """Make sure network metadata entity changes update the trie index."""
        self.model = PreseemNetworkMetadataModel(self.client, refs_by_type=True)
        self.client.nms = [
            _make_entity('192.0.2.1'),
            _make_entity('192.0.2.54'),
            _make_entity('192.0.2.52/30'),
        ]
        self.update()

        # Check results of initial load
        self.assertEqual(len(self.model.nms), 3)
        self.assertEqual(len(self.model.nms), len(self.model.pt4))
        key = self.model.pt4.get_key('192.0.2.54')
        self.assertEqual(key, '192.0.2.54/32')
        par = self.model.pt4.parent(key)
        self.assertEqual(par, '192.0.2.52/30')

        # Add another entity
        self.set_nm(_make_entity('192.0.2.2'))
        self.assertEqual(len(self.model.nms), 4)
        self.assertEqual(len(self.model.nms), len(self.model.pt4))

        # Delete an entity
        self.del_nm(_make_entity('192.0.2.54'))
        self.assertEqual(len(self.model.nms), 3)
        key = self.model.pt4.get_key('192.0.2.54')
        self.assertEqual(key, '192.0.2.52/30')


def _make_ref_msg(ref, delete=False):
    """Helper to create a protobuf ref subscription message."""
    ts = network_metadata_pb2.google_dot_protobuf_dot_timestamp__pb2.Timestamp()
    ts.FromNanoseconds(int(time.time() * 1000000000))
    if ref is None:
        action = network_metadata_pb2.NetworkMetadataReferenceEntityAction.LOADED
        entity = None
    else:
        action = (network_metadata_pb2.NetworkMetadataReferenceEntityAction.DELETE
                  if delete else
                  network_metadata_pb2.NetworkMetadataReferenceEntityAction.SET)
        entity = network_metadata_pb2.NetworkMetadataReferenceEntity(
            type=ref.type, value=ref.value)
        preseem.grpc_nm_model._nm_set_attrs(entity.attributes, ref.attributes)
    return network_metadata_pb2.NetworkMetadataReferenceEntityAction(
        action=action, entity=entity, timestamp=ts)


REF_A = NetworkMetadataReference('service', 'svc1', {'username': 'Alice'})
REF_B = NetworkMetadataReference('service', 'svc2', {'username': 'Bob'})
REF_C = NetworkMetadataReference('device', 'dev1', {'hostname': 'router1'})


class TestRefHandler(unittest.TestCase):
    """Test ref_handler generator in isolation (no asyncio needed)."""

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        client = FakeMetadataClient()
        self.model = PreseemNetworkMetadataModel(client, refs_by_type=True)

    def tearDown(self):
        self.model.close()
        self.loop.close()

    def test_initial_sync_batches_refs(self):
        """During initial sync, SET messages are accumulated and yielded as a single SYN op."""
        msgs = [_make_ref_msg(REF_A), _make_ref_msg(REF_B), _make_ref_msg(None)]
        it = GrpcIterator(msgs)

        ops = _consume_handler(self.model, it)

        self.assertEqual(len(ops), 1)
        op, syn_dict = ops[0]
        self.assertEqual(op, Op.SYN)
        self.assertIn('service', syn_dict)
        self.assertEqual(len(syn_dict['service']), 2)
        self.assertEqual(syn_dict['service']['svc1'].attributes['username'], 'Alice')
        self.assertEqual(syn_dict['service']['svc2'].attributes['username'], 'Bob')

    def test_initial_sync_multiple_types(self):
        """SYN dict is keyed by type."""
        msgs = [_make_ref_msg(REF_A), _make_ref_msg(REF_C), _make_ref_msg(None)]
        it = GrpcIterator(msgs)

        ops = _consume_handler(self.model, it)

        self.assertEqual(len(ops), 1)
        _, syn_dict = ops[0]
        self.assertIn('service', syn_dict)
        self.assertIn('device', syn_dict)
        self.assertEqual(syn_dict['device']['dev1'].attributes['hostname'], 'router1')

    def test_post_sync_set_yields_immediately(self):
        """After LOADED, SET messages yield individual Op.SET ops."""
        msgs = [_make_ref_msg(REF_A), _make_ref_msg(None), _make_ref_msg(REF_B)]
        it = GrpcIterator(msgs)

        ops = _consume_handler(self.model, it)

        self.assertEqual(len(ops), 2)
        self.assertEqual(ops[0][0], Op.SYN)
        self.assertEqual(ops[1][0], Op.SET)
        self.assertEqual(ops[1][1].value, 'svc2')

    def test_post_sync_delete_yields_immediately(self):
        """After LOADED, DELETE messages yield individual Op.DEL ops."""
        msgs = [_make_ref_msg(REF_A), _make_ref_msg(None),
                _make_ref_msg(REF_A, delete=True)]
        it = GrpcIterator(msgs)

        ops = _consume_handler(self.model, it)

        self.assertEqual(len(ops), 2)
        self.assertEqual(ops[0][0], Op.SYN)
        self.assertEqual(ops[1][0], Op.DEL)
        self.assertEqual(ops[1][1].value, 'svc1')

    def test_empty_sync(self):
        """A LOADED with no preceding SETs yields an empty SYN dict."""
        it = GrpcIterator([_make_ref_msg(None)])

        ops = _consume_handler(self.model, it)

        self.assertEqual(len(ops), 1)
        self.assertEqual(ops[0][0], Op.SYN)
        self.assertEqual(ops[0][1], {})

    def test_duplicate_ref_during_sync_keeps_last(self):
        """If the same (type, value) appears twice in sync, last one wins."""
        ref_a2 = NetworkMetadataReference('service', 'svc1', {'username': 'Alice2'})
        msgs = [_make_ref_msg(REF_A), _make_ref_msg(ref_a2), _make_ref_msg(None)]
        it = GrpcIterator(msgs)

        ops = _consume_handler(self.model, it)
        syn_dict = ops[0][1]
        self.assertEqual(syn_dict['service']['svc1'].attributes['username'], 'Alice2')


class TestHandleRefUpdate2(unittest.TestCase):
    """Test handle_ref_update2 async callback in isolation."""

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.client = FakeMetadataClient()
        self.model = PreseemNetworkMetadataModel(self.client, refs_by_type=True)

    def tearDown(self):
        self.model.close()
        self.loop.close()

    def _run(self, coro):
        return self.loop.run_until_complete(coro)

    def test_syn_populates_refs(self):
        """Op.SYN replaces the refs dict."""
        syn_dict = {
            'service': {
                'svc1': REF_A,
                'svc2': REF_B,
            }
        }
        self._run(self.model.handle_ref_update2(Op.SYN, syn_dict))

        self.assertIs(self.model.refs, syn_dict)
        self.assertIn('service', self.model.refs_last_updated)

    def test_syn_notifies_waiters(self):
        """Op.SYN notifies the _refs_loaded condition."""
        notified = []

        async def wait_and_record():
            async with self.model._refs_loaded:
                await self.model._refs_loaded.wait()
                notified.append(True)

        async def run():
            task = self.loop.create_task(wait_and_record())
            await asyncio.sleep(0)  # let the waiter start
            await self.model.handle_ref_update2(Op.SYN, {})
            await asyncio.sleep(0)  # let the waiter finish
            return task

        task = self._run(run())
        self.assertTrue(notified)

    def test_set_adds_ref(self):
        """Op.SET adds a ref to the model."""
        self._run(self.model.handle_ref_update2(Op.SYN, {}))
        self._run(self.model.handle_ref_update2(Op.SET, REF_A))

        self.assertEqual(self.model.refs['service']['svc1'], REF_A)

    def test_set_updates_existing_ref(self):
        """Op.SET replaces an existing ref with the same key."""
        self._run(self.model.handle_ref_update2(Op.SYN, {}))
        self._run(self.model.handle_ref_update2(Op.SET, REF_A))

        updated = NetworkMetadataReference('service', 'svc1', {'username': 'Alice2'})
        self._run(self.model.handle_ref_update2(Op.SET, updated))

        self.assertEqual(self.model.refs['service']['svc1'].attributes['username'], 'Alice2')

    def test_del_removes_ref(self):
        """Op.DEL removes a ref from the model."""
        syn = {'service': {'svc1': REF_A, 'svc2': REF_B}}
        self._run(self.model.handle_ref_update2(Op.SYN, syn))
        self._run(self.model.handle_ref_update2(Op.DEL, REF_A))

        self.assertNotIn('svc1', self.model.refs.get('service', {}))
        self.assertIn('svc2', self.model.refs['service'])

    def test_del_removes_empty_type_dict(self):
        """Op.DEL cleans up the type dict when it becomes empty."""
        syn = {'service': {'svc1': REF_A}}
        self._run(self.model.handle_ref_update2(Op.SYN, syn))
        self._run(self.model.handle_ref_update2(Op.DEL, REF_A))

        self.assertNotIn('service', self.model.refs)

    def test_del_nonexistent_ref_is_noop(self):
        """Op.DEL for a ref that doesn't exist does nothing."""
        self._run(self.model.handle_ref_update2(Op.SYN, {}))
        self._run(self.model.handle_ref_update2(Op.DEL, REF_A))  # should not raise
        self.assertEqual(self.model.refs, {})

    def test_set_with_callback(self):
        """Op.SET invokes the user callback."""
        received = []

        async def cbk(ref, deleted=False):
            received.append((ref, deleted))

        self.model._cbk = cbk
        self._run(self.model.handle_ref_update2(Op.SYN, {}))
        self._run(self.model.handle_ref_update2(Op.SET, REF_A))

        self.assertEqual(len(received), 1)
        self.assertEqual(received[0], (REF_A, False))

    def test_del_with_callback(self):
        """Op.DEL invokes the user callback with deleted=True."""
        received = []

        async def cbk(ref, deleted=False):
            received.append((ref, deleted))

        self.model._cbk = cbk
        syn = {'service': {'svc1': REF_A}}
        self._run(self.model.handle_ref_update2(Op.SYN, syn))
        received.clear()  # ignore SYN callbacks
        self._run(self.model.handle_ref_update2(Op.DEL, REF_A))

        self.assertEqual(len(received), 1)
        self.assertEqual(received[0][1], True)

    def test_syn_with_callback(self):
        """Op.SYN invokes the callback for each ref."""
        received = []

        async def cbk(ref, deleted=False):
            received.append(ref)

        self.model._cbk = cbk
        syn = {'service': {'svc1': REF_A, 'svc2': REF_B}}
        self._run(self.model.handle_ref_update2(Op.SYN, syn))

        self.assertEqual(len(received), 2)
        values = {r.value for r in received}
        self.assertEqual(values, {'svc1', 'svc2'})


class TestRefHandlerWithIndex(unittest.TestCase):
    """Test that indexes are maintained through the new handler path."""

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.client = FakeMetadataClient()
        indexes = [NetworkMetadataLowercaseIndex('service', 'username')]
        self.model = PreseemNetworkMetadataModel(
            self.client, refs_by_type=True, indexes=indexes)

    def tearDown(self):
        self.model.close()
        self.loop.close()

    def _run(self, coro):
        return self.loop.run_until_complete(coro)

    def test_syn_builds_index(self):
        """Op.SYN rebuilds indexes from the full ref set."""
        syn = {'service': {'svc1': REF_A, 'svc2': REF_B}}
        self._run(self.model.handle_ref_update2(Op.SYN, syn))

        refs = self.model.get_ref_index('service', 'username', 'alice') or {}
        self.assertEqual(len(refs), 1)
        self.assertIn('svc1', refs)

    def test_set_updates_index(self):
        """Op.SET adds to the index."""
        self._run(self.model.handle_ref_update2(Op.SYN, {}))
        self._run(self.model.handle_ref_update2(Op.SET, REF_A))

        refs = self.model.get_ref_index('service', 'username', 'alice') or {}
        self.assertEqual(len(refs), 1)

    def test_set_update_moves_index(self):
        """Updating a ref's indexed attribute moves it in the index."""
        self._run(self.model.handle_ref_update2(Op.SYN, {}))
        self._run(self.model.handle_ref_update2(Op.SET, REF_A))

        updated = NetworkMetadataReference('service', 'svc1', {'username': 'Alicia'})
        self._run(self.model.handle_ref_update2(Op.SET, updated))

        old_refs = self.model.get_ref_index('service', 'username', 'alice') or {}
        self.assertEqual(len(old_refs), 0)

        new_refs = self.model.get_ref_index('service', 'username', 'alicia') or {}
        self.assertEqual(len(new_refs), 1)

    def test_del_removes_from_index(self):
        """Op.DEL removes from the index."""
        syn = {'service': {'svc1': REF_A}}
        self._run(self.model.handle_ref_update2(Op.SYN, syn))
        self._run(self.model.handle_ref_update2(Op.DEL, REF_A))

        refs = self.model.get_ref_index('service', 'username', 'alice') or {}
        self.assertEqual(len(refs), 0)


class TestRefIntegration(unittest.TestCase):
    """Integration test: ref_handler + handle_ref_update2 wired together via threading,
    simulating the real Subscription pattern."""

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.client = FakeMetadataClient()
        indexes = [NetworkMetadataLowercaseIndex('service', 'username')]
        self.model = PreseemNetworkMetadataModel(
            self.client, refs_by_type=True, indexes=indexes)

    def tearDown(self):
        self.model.close()
        self.loop.run_until_complete(self.loop.shutdown_asyncgens())
        self.loop.close()

    def _run_handler_on_messages(self, messages):
        """Run ref_handler in a thread, dispatching ops to handle_ref_update2 on the event loop."""
        it = GrpcIterator(messages)
        loop = self.loop

        def worker():
            try:
                for op in self.model.ref_handler(it):
                    fut = asyncio.run_coroutine_threadsafe(
                        self.model.handle_ref_update2(*op), loop)
                    fut.result(timeout=5)
            except _Disconnected:
                pass  # iterator exhausted, simulates gRPC disconnect

        async def run():
            await loop.run_in_executor(None, worker)

        self.loop.run_until_complete(run())

    def test_full_sync_and_live_updates(self):
        """Full flow: initial sync, then live SET and DEL."""
        messages = [
            _make_ref_msg(REF_A),
            _make_ref_msg(REF_B),
            _make_ref_msg(None),     # LOADED
            _make_ref_msg(REF_C),    # live SET
            _make_ref_msg(REF_A, delete=True),  # live DEL
        ]
        self._run_handler_on_messages(messages)

        # After sync: svc1 and svc2 loaded
        # After live: dev1 added, svc1 deleted
        self.assertNotIn('svc1', self.model.refs.get('service', {}))
        self.assertEqual(self.model.refs['service']['svc2'], REF_B)
        self.assertEqual(self.model.refs['device']['dev1'], REF_C)

        # Index should reflect final state
        refs = self.model.get_ref_index('service', 'username', 'alice') or {}
        self.assertEqual(len(refs), 0)
        refs = self.model.get_ref_index('service', 'username', 'bob') or {}
        self.assertEqual(len(refs), 1)

    def test_large_sync(self):
        """Sync with many refs to verify batching works."""
        refs = [NetworkMetadataReference('service', f'svc{i}', {'username': f'user{i}'})
                for i in range(500)]
        messages = [_make_ref_msg(r) for r in refs] + [_make_ref_msg(None)]

        self._run_handler_on_messages(messages)

        self.assertEqual(len(self.model.refs['service']), 500)

    def test_end_to_end_via_fake_client(self):
        """Full end-to-end using FakeMetadataClient.nm_ref_subscribe -> update()."""
        self.client.refs = [REF_A, REF_B, REF_C]
        self.loop.run_until_complete(self.model.update())

        self.assertEqual(len(self.model.refs['service']), 2)
        self.assertEqual(len(self.model.refs['device']), 1)
        refs = self.model.get_ref_index('service', 'username', 'bob') or {}
        self.assertEqual(len(refs), 1)



"""
Load Balancer Manager GRPC integration for requesting URL map syncs
"""
from preseem_protobuf.load_balancer_manager.url_map_pb2 import SyncUrlMapRequest
from preseem_protobuf.load_balancer_manager.url_map_pb2_grpc import UrlMapServiceStub
from preseem import PreseemGrpcClient


class LoadBalancerManagerClient:
    """Client for requesting URL map syncs."""

    def __init__(self, api_key, host, port):
        self._client = PreseemGrpcClient(api_key=api_key, host=host, port=port)
        self._svc = None

    def sync_url_map(self):
        if not self._client:
            return
        if not self._svc:
            self._client._connect()
            self._svc = UrlMapServiceStub(self._client._channel)
        return self._svc.SyncUrlMap(SyncUrlMapRequest())



"""
Implement counter-to-interval conversion and interpolation as specified in STM-6055.

In PPA-1575 we added explicit attributes to the Element message in the update to provide
a way to signal a change of counter source on an element, to reset the counters on a
counter source change.  This works as follows:
  If the counter source of the main element changes, its interface stats are reset.
  If the counter source of a peer changes, its interface stats are reset, and all
  links between the main element and that peer have their counters reset, too.
"""
from collections import namedtuple
import logging  # PPA-1575 we may remove logging after we exercise this a bit.
import time

from preseem_protobuf.network_poller import network_poller_pb2

interface_counters = [
    "in_octets",
    "in_ucast_pkts",
    "in_nucast_pkts",
    "in_discards",
    "in_errors",
    "in_unknown_protos",
    "out_octets",
    "out_ucast_pkts",
    "out_nucast_pkts",
    "out_discards",
    "out_errors",
]

link_counters = [
    "in_octets",
    "in_ucast_pkts",
    "in_nucast_pkts",
    "in_discards",
    "in_errors",
    "in_unknown_protos",
    "out_octets",
    "out_ucast_pkts",
    "out_nucast_pkts",
    "out_discards",
    "out_errors",
    "out_retransmits",
]


def process_msg_to_send(msg):
    # HS-85 Sometimes we don't have aggregate "port-level" metrics on the parent side
    # of a set of PTMP links.  In that case, we can populate this with the sum of the
    # link-level metrics.
    for intf in list(
            msg.data.interfaces) + [i for m in msg.data.modules for i in m.interfaces]:
        if len(intf.links) == 0:
            continue
        for field in interface_counters:
            if not intf.HasField(field):
                link_fields = [
                    getattr(l, field) for l in intf.links if l.HasField(field)
                ]
                if link_fields:
                    setattr(intf, field, sum(link_fields))
    return msg


def process_element(cur_inc, cur_int, new_inc, new_int, int_factor, reset_counters,
                    reset_links):
    """Update interval counts given previous state."""
    cur_inc_interfaces = {
        x.poller_hash: x
        for x in list(cur_inc.interfaces) +
        [i for m in cur_inc.modules for i in m.interfaces]
    } if cur_inc else {}
    cur_int_interfaces = {
        x.poller_hash: x
        for x in list(cur_int.interfaces) +
        [i for m in cur_int.modules for i in m.interfaces]
    } if cur_int else {}
    new_int_interfaces = {
        x.poller_hash: x
        for x in list(new_int.interfaces) +
        [i for m in new_int.modules for i in m.interfaces]
    }
    for new_inc_intf in list(
            new_inc.interfaces) + [i for m in new_inc.modules for i in m.interfaces]:
        cur_inc_intf = cur_inc_interfaces.get(new_inc_intf.poller_hash)
        cur_int_intf = cur_int_interfaces.get(new_inc_intf.poller_hash)
        new_int_intf = new_int_interfaces.get(new_inc_intf.poller_hash)
        for field in interface_counters:
            # If either message doesn't have the field, both intervals get nulled out.
            # The next one after should have data.
            counter1 = counter2 = None
            if cur_inc_intf and cur_inc_intf.HasField(field):
                counter1 = getattr(cur_inc_intf, field)
            if new_inc_intf.HasField(field):
                counter2 = getattr(new_inc_intf, field)
            if counter1 is not None and counter2 is not None and counter1 <= counter2 and not reset_counters:
                delta = counter2 - counter1
                val0 = int(delta * int_factor)
                val1 = delta - val0
                if cur_int_intf.HasField(field):
                    setattr(cur_int_intf, field, getattr(cur_int_intf, field) + val0)
                if new_int_intf.HasField(field):
                    setattr(new_int_intf, field, val1)
            else:
                # We don't have both counter values, or the counter reset.
                # Treat this as a counter reset either way.
                if counter1 and counter2 == 0:
                    # special case where we got a zero counter: this _could_
                    # be a polling error and not a reset on the device side;
                    # we will still set the current message to null but we
                    # copy the previous value across in case the counters
                    # continue where they left off.  If it was in fact a
                    # reset, we'll handle the reset when we get a non-zero
                    # value.
                    setattr(new_inc_intf, field, counter1)
                if cur_int_intf:
                    cur_int_intf.ClearField(field)
                new_int_intf.ClearField(field)

        cur_inc_links = {
            x.poller_hash: x
            for x in cur_inc_intf.links
        } if cur_inc_intf else {}
        cur_int_links = {
            x.poller_hash: x
            for x in cur_int_intf.links
        } if cur_int_intf else {}
        new_int_links = {x.poller_hash: x for x in new_int_intf.links}
        for new_inc_link in new_inc_intf.links:
            cur_inc_link = cur_inc_links.get(new_inc_link.poller_hash)
            cur_int_link = cur_int_links.get(new_inc_link.poller_hash)
            new_int_link = new_int_links.get(new_inc_link.poller_hash)
            for field in link_counters:
                counter1 = counter2 = None
                if cur_inc_link and cur_inc_link.HasField(field):
                    counter1 = getattr(cur_inc_link, field)
                if new_inc_link.HasField(field):
                    counter2 = getattr(new_inc_link, field)
                if counter1 is not None and counter2 is not None and counter1 <= counter2 and (
                        not reset_links or new_inc_link.poller_hash not in reset_links):
                    delta = counter2 - counter1
                    val0 = int(delta * int_factor)
                    val1 = delta - val0
                    if cur_int_link and cur_int_link.HasField(field):
                        setattr(cur_int_link, field,
                                getattr(cur_int_link, field) + val0)
                    if new_int_link.HasField(field):
                        setattr(new_int_link, field, val1)
                else:
                    # We don't have both counter values, or the counter reset.
                    # Treat this as a counter reset either way.
                    if counter1 and counter2 == 0:
                        setattr(new_inc_link, field, counter1)
                    if cur_int_link:
                        cur_int_link.ClearField(field)
                    new_int_link.ClearField(field)


class UpdateIntervalManager:
    """Class responsible for counter->interval conversion"""
    State = namedtuple('State', ('orig_msg', 'new_msg', 'time'))

    def __init__(self):
        self.elem_state = {}

    def handle_message(self, msg, id_key=None):
        """Handle a message from the network_poller_grpc post method."""
        if not msg.time or not msg.data.poller_hash:
            return msg  # only apply this logic to messages with data.

        # we use a unique key from the message as well as an (optional) id_key
        # passed in.  The id_key lets us disambiguate cases where the poller
        # sends two messages for an element (e.g. it's entered as an AP but
        # is actually a CPE radio of another AP).
        elem_id = (msg.data.poller_hash, id_key)
        cur = self.elem_state.get(elem_id) or self.State(None, None, None)

        interval_end = msg.time.seconds - msg.time.seconds % 60 + 60
        interval_start = cur.new_msg.interval_end.seconds if cur.new_msg else interval_end - 60
        new = self.State(msg, network_poller_pb2.NetworkElementUpdate(),
                         time.monotonic())
        new.new_msg.CopyFrom(msg)
        new.new_msg.interval_start.FromSeconds(interval_start)
        new.new_msg.interval_end.FromSeconds(interval_end)

        delta_t = msg.time.seconds - cur.new_msg.time.seconds if cur.new_msg else None
        if delta_t is not None and delta_t <= 0.0:
            raise RuntimeError('Timestamp did not increment')
        int_factor = None if cur.new_msg is None else (
            cur.new_msg.interval_end.seconds -
            cur.new_msg.time.seconds) / (msg.time.seconds - cur.new_msg.time.seconds)
        if cur.new_msg and cur.new_msg.interval_end == new.new_msg.interval_end:
            # The new message is within the interval of the current message.
            # Update the current message in progress and keep the new message.
            process_element(cur.orig_msg.data, cur.new_msg.data, msg.data,
                            new.new_msg.data, 1.0, None, None)

            cur.new_msg.time.FromSeconds(msg.time.seconds)
            new = self.State(msg, cur.new_msg, new.time)
            cur = self.State(None, None, None)
        else:
            reset_counters = cur.orig_msg.data.counter_source != msg.data.counter_source if cur.orig_msg else False
            if reset_counters:
                logging.info("Reset counter for %s: %s -> %s", msg.data.management_ip,
                             cur.orig_msg.data.counter_source, msg.data.counter_source)
            cur_inc_peers = {
                x.poller_hash: x
                for x in cur.orig_msg.peers
            } if cur.orig_msg else {}
            cur_int_peers = {
                x.poller_hash: x
                for x in cur.new_msg.peers
            } if cur.new_msg else {}

            reset_main_links = set()  # Links to peers to reset for main element
            reset_peer_links = set()  # Links to main element to reset for peers
            for peer in msg.peers:
                cur_inc_peer = cur_inc_peers.get(peer.poller_hash)
                if cur_inc_peer and cur_inc_peer.counter_source != peer.counter_source:
                    for intf in list(peer.interfaces) + [
                            i for m in peer.modules for i in m.interfaces
                    ]:
                        if len(intf.links) > 0:
                            reset_main_links.add(intf.poller_hash)
                    if not reset_peer_links:  # do this once if any peers are resetting
                        for intf in list(msg.data.interfaces) + [
                                i for m in msg.data.modules for i in m.interfaces
                        ]:
                            if len(intf.links) > 0:
                                reset_peer_links.add(intf.poller_hash)

            process_element(cur.orig_msg.data if cur.orig_msg else None,
                            cur.new_msg.data if cur.new_msg else None, msg.data,
                            new.new_msg.data, int_factor, reset_counters,
                            reset_main_links)
            new_int_peers = {x.poller_hash: x for x in new.new_msg.peers}
            for peer in msg.peers:
                cur_inc_peer = cur_inc_peers.get(peer.poller_hash)
                cur_int_peer = cur_int_peers.get(peer.poller_hash)
                new_int_peer = new_int_peers.get(peer.poller_hash)
                reset_counters = cur_inc_peer and cur_inc_peer.counter_source != peer.counter_source
                if reset_counters:
                    logging.info("Reset counter for %s[%s]: %s -> %s",
                                 msg.data.management_ip, peer.system_mac_address,
                                 cur.orig_msg.data.counter_source,
                                 msg.data.counter_source)
                process_element(cur_inc_peer, cur_int_peer, peer, new_int_peer,
                                int_factor, reset_counters,
                                reset_peer_links if reset_counters else None)

        self.elem_state.pop(elem_id, None)  # keep the dict in LRU order
        self.elem_state[elem_id] = new
        return process_msg_to_send(cur.new_msg) if cur.new_msg else None

    def timeout_messages(self, expire_time):
        """Remove any messages we are holding that are older than expire_time seconds."""
        now = time.monotonic()
        dels = []
        for elem_id, state in self.elem_state.items():
            # see how old time is.  If expired, delete it.
            if state.time + expire_time < now:
                dels.append(elem_id)
            else:
                break  # we're past any expired records in the dict
        for elem_id in dels:
            del self.elem_state[elem_id]
